/* -*- Mode: C; tab-width: 8; indent-tabs-mode: t; c-basic-offset: 8 -*- */
/*  This file is part of MixMagic.

    Copyright (C) 2000-2001 Jonas Borgstrm <jonas@bengburken.net>

    This program is free software; you can redistribute it and/or
    modify it under the terms of the GNU General Public
    License as published by the Free Software Foundation; either
    version 2 of the License, or (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the 
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software 
    Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
*/

#include "mixmagic.h"
#include "application.h"
#include "block.h"
#include "loopblock.h"
#include "mixer.h"
#include "track.h"
#include "plugins.h"
#include <math.h>

#define RADIUS 3
#define MIN_DISTANCE 8
#define MIN_Y_DISTANCE 0.1
#define SNAP_RESISTANCE (prefs->snap_resistance)

GdkAtom block_atom = GDK_NONE;

//static GtkVBoxClass *parent_class = NULL;
static GtkDrawingAreaClass *parent_class = NULL;
static gint block_expose (GtkWidget      *widget, GdkEventExpose *event);
static void block_class_init (BlockClass *class);
static void block_init (Block *block);
static gboolean block_configure (GtkWidget *widget,GdkEventConfigure *event);
static gint block_motion_notify (GtkWidget *widget, GdkEventMotion *event);
static gint block_button_press (GtkWidget *widget, GdkEventButton *event);
static gint block_button_release (GtkWidget *widget, GdkEventButton *event);
static void block_finalize (GtkObject *object);
static void block_realize (GtkWidget *widget);
static void block_unrealize (GtkWidget *widget);
static gint block_selection_clear( GtkWidget *widget, GdkEventSelection *event);
static void block_selection_get (GtkWidget *widget, GtkSelectionData *selection_data, guint info, guint time_stamp);
static void block_create_popup_menu (Block *block);
void block_mix_into_buffer (Block *block, gint16 *buffer, gint position, gint length);
static void mix_into_buffer (Block *block, gint16 *buffer, gint position, gint length);
static void lower_left_box_motion (Block *block, GdkEventMotion *event, gint x, gint y);
static void lower_right_box_motion (Block *block, GdkEventMotion *event, gint x, gint y);
static void lower_half_motion (Block *block, GdkEventMotion *event, gint x, gint y);
static void upper_half_motion (Block *block, GdkEventMotion *event, gint x, gint y);

enum {
        MIX_INTO_BUFFER_SIGNAL,
	LOWER_LEFT_BOX_MOTION,
	LOWER_RIGHT_BOX_MOTION,
	LOWER_HALF_MOTION,
	UPPER_HALF_MOTION,
        LAST_SIGNAL
};

static gint block_signals[LAST_SIGNAL] = { 0 };

guint
block_get_type ()
{
        static guint block_type = 0;
	
        if (!block_type) {

		GtkTypeInfo block_info = {
			"Block",
			sizeof (Block),
			sizeof (BlockClass),
			(GtkClassInitFunc) block_class_init,
			(GtkObjectInitFunc) block_init,
			(GtkArgSetFunc) NULL,
			(GtkArgGetFunc) NULL
		};
	    
		block_type = gtk_type_unique (gtk_drawing_area_get_type (), &block_info);
          }
	
        return block_type;
}

static void
block_class_init (BlockClass *class)
{
	GtkWidgetClass *widget_class;
	GtkObjectClass *object_class;

	parent_class = gtk_type_class (GTK_TYPE_DRAWING_AREA);

	object_class = (GtkObjectClass*) class;
	object_class->finalize = block_finalize;

	block_signals[MIX_INTO_BUFFER_SIGNAL] = gtk_signal_new ("mix_into_buffer",
								GTK_RUN_LAST,
								object_class->type,
								GTK_SIGNAL_OFFSET (BlockClass, mix_into_buffer),
								gtk_marshal_NONE__POINTER_INT_INT, 
								GTK_TYPE_NONE, 3,
								GTK_TYPE_POINTER, GTK_TYPE_INT,	GTK_TYPE_INT);

	block_signals[LOWER_LEFT_BOX_MOTION] = gtk_signal_new ("lower_left_box_motion",
							       GTK_RUN_LAST,
							       object_class->type,
							       GTK_SIGNAL_OFFSET (BlockClass, lower_left_box_motion),
							       gtk_marshal_NONE__POINTER_INT_INT, 
							       GTK_TYPE_NONE, 3,
							       GTK_TYPE_POINTER, GTK_TYPE_INT,	GTK_TYPE_INT);
	
	block_signals[LOWER_RIGHT_BOX_MOTION] = gtk_signal_new ("lower_right_box_motion",
								GTK_RUN_LAST,
								object_class->type,
								GTK_SIGNAL_OFFSET (BlockClass, lower_right_box_motion),
								gtk_marshal_NONE__POINTER_INT_INT, 
								GTK_TYPE_NONE, 3,
								GTK_TYPE_POINTER, GTK_TYPE_INT,	GTK_TYPE_INT);

	block_signals[LOWER_HALF_MOTION] = gtk_signal_new ("lower_half_motion",
							   GTK_RUN_LAST,
							   object_class->type,
							   GTK_SIGNAL_OFFSET (BlockClass, lower_half_motion),
							   gtk_marshal_NONE__POINTER_INT_INT, 
							   GTK_TYPE_NONE, 3,
							   GTK_TYPE_POINTER, GTK_TYPE_INT,	GTK_TYPE_INT);

	block_signals[UPPER_HALF_MOTION] = gtk_signal_new ("upper_half_motion",
							   GTK_RUN_LAST,
							   object_class->type,
							   GTK_SIGNAL_OFFSET (BlockClass, upper_half_motion),
							   gtk_marshal_NONE__POINTER_INT_INT, 
							   GTK_TYPE_NONE, 3,
							   GTK_TYPE_POINTER, GTK_TYPE_INT, GTK_TYPE_INT);

        gtk_object_class_add_signals (object_class, block_signals, LAST_SIGNAL);

	class->mix_into_buffer = mix_into_buffer;
	class->lower_left_box_motion = lower_left_box_motion;
	class->lower_right_box_motion = lower_right_box_motion;
	class->lower_half_motion = lower_half_motion;
	class->upper_half_motion = upper_half_motion;

	widget_class = (GtkWidgetClass*) class;
	widget_class->expose_event = block_expose;
	widget_class->configure_event = block_configure;
	widget_class->motion_notify_event = block_motion_notify;
	widget_class->button_press_event = block_button_press;
	widget_class->button_release_event = block_button_release;
	widget_class->realize = block_realize;
	widget_class->unrealize = block_unrealize;
	widget_class->selection_clear_event = block_selection_clear;
	widget_class->selection_get = block_selection_get;
}

static void
block_init (Block *block)
{
        /* Get the atom corresponding to the string "Block" */
        if (block_atom == GDK_NONE)
		block_atom = gdk_atom_intern ("Block", FALSE);	

	gtk_selection_add_target (GTK_WIDGET (block),
                                  GDK_SELECTION_PRIMARY,
                                  block_atom,
                                  1);

	gtk_widget_set_events (GTK_WIDGET (block), 
			       GDK_POINTER_MOTION_HINT_MASK 
			       | GDK_POINTER_MOTION_MASK 
			       | GDK_BUTTON_PRESS_MASK
			       | GDK_BUTTON_RELEASE_MASK);

	block->pixmap = NULL;
	block->click_flag = FALSE;
	block->selection_flag = FALSE;

	block->envelope_mode = NO_ENVELOPE;

	block_create_popup_menu (block);
	block->cursor_type = GDK_TOP_LEFT_ARROW;
}

static void
block_realize (GtkWidget *widget)
{
	gint color_index;
	GdkColor selection_color = {1, 0x0000, 0x0000, 0xffff};
	GdkColor amp_color = {1, 0x7777, 0x7777, 0xffff};
	GdkColor pan_color = {1, 0xffff, 0x7777, 0x7777};

#define NUM_WAV_COLORS (sizeof (wav_colors) / sizeof (GdkColor))

	GdkColor wav_colors[] = {{1, 0x7777, 0x0000, 0xffff},
				 {1, 0x0000, 0xffff, 0x0000},
				 {1, 0xffff, 0x7777, 0x0000},
				 {1, 0xffff, 0xffff, 0x0000},
				 {1, 0xffff, 0x0000, 0xffff},
				 {1, 0x0000, 0xffff, 0xffff},
				 {1, 0x0000, 0x0000, 0xaaaa},
				 {1, 0x0000, 0xaaaa, 0x0000},
				 {1, 0xaaaa, 0x0000, 0x0000},
				 {1, 0x7777, 0x7777, 0x0000},
				 {1, 0x0000, 0x7777, 0x7777}
	};

	Block *block = BLOCK (widget);

	g_return_if_fail (widget != NULL);
	g_return_if_fail (IS_BLOCK (widget));

	(* GTK_WIDGET_CLASS (parent_class)->realize) (widget);

	/* Create GC */
	if (block->wav_gc == NULL && block->sample) {
		block->wav_gc = gdk_gc_new (GTK_WIDGET(block)->window);

		color_index = (*( (gint *)(block->sample->map_addr + 
					   block->sample->map_size / 2))) % NUM_WAV_COLORS;
	
		/* Stupid hack to get a custom color for every sample */
		gdk_color_alloc (gtk_widget_get_colormap (widget), 
				  &wav_colors [color_index]);

		gdk_gc_set_foreground(block->wav_gc, &wav_colors [color_index]);
	}

	/* Create selection GC */
	if (block->selection_gc == NULL) {
		block->selection_gc = gdk_gc_new (GTK_WIDGET(block)->window);
	
		gdk_color_alloc (gtk_widget_get_colormap (widget), &selection_color);
		gdk_gc_set_background (block->selection_gc, &selection_color);
		gdk_gc_set_function (block->selection_gc, GDK_INVERT);
	}
	/* Create pan GC */
	if (block->pan_gc == NULL) {
		block->pan_gc = gdk_gc_new (GTK_WIDGET(block)->window);
	
		gdk_color_alloc (gtk_widget_get_colormap (widget), &pan_color);
		gdk_gc_set_foreground (block->pan_gc, &pan_color);
	}
	/* Create Amp GC */
	if (block->amp_gc == NULL) {
		block->amp_gc = gdk_gc_new (GTK_WIDGET(block)->window);
	
		gdk_color_alloc (gtk_widget_get_colormap (widget), &amp_color);
		gdk_gc_set_foreground (block->amp_gc, &amp_color);
	}
}

static void
block_unrealize (GtkWidget *widget)
{
	Block *block = BLOCK (widget);
	
	g_return_if_fail (widget != NULL);
	g_return_if_fail (IS_BLOCK (widget));
		
	(* GTK_WIDGET_CLASS (parent_class)->unrealize) (widget);

	if (block->wav_gc) {
		gdk_gc_unref (block->wav_gc);
		block->wav_gc = NULL;
	}
	if (block->pan_gc) {
		gdk_gc_unref (block->pan_gc);
		block->pan_gc = NULL;
	}
	if (block->amp_gc) {
		gdk_gc_unref (block->amp_gc);
		block->amp_gc = NULL;
	}
	if (block->selection_gc) {
		gdk_gc_unref (block->selection_gc);
		block->selection_gc = NULL;
	}
}

static void
free_envelope (gpointer data, gpointer user_data)
{
	g_free (data);
}

static void
block_finalize (GtkObject *object)
{
	Block *block = BLOCK(object);

	g_return_if_fail (object != NULL);
	g_return_if_fail (IS_BLOCK(object));

	g_print ("block finalize\n");

	if (block->pixmap)
		gdk_pixmap_unref (block->pixmap);

	if (block->wav_gc)
		gdk_gc_unref (block->wav_gc);

	if (block->pan_gc)
		gdk_gc_unref (block->wav_gc);

	if (block->amp_gc)
		gdk_gc_unref (block->amp_gc);

	if (block->selection_gc)
		gdk_gc_unref (block->selection_gc);

	if (block->popup_menu)
		gtk_widget_destroy (block->popup_menu);

	/* Free amplitude envelope memory */
	g_list_foreach (block->amp_env, free_envelope, NULL);
	g_list_free (block->amp_env);
	/* Free pan envelope memory */
	g_list_foreach (block->pan_env, free_envelope, NULL);
	g_list_free (block->pan_env);
	
	if (block->sample)
		sample_unref (block->sample);
	
	(* GTK_OBJECT_CLASS (parent_class)->finalize) (object);
}

static gint
envelope_compare_func (gconstpointer data1, gconstpointer data2)
{
	const Envelope *e1 = (Envelope *)data1;
	const Envelope *e2 = (Envelope *)data2;

	if (e1->x == e2->x)
		return 0;
	else if (e1->x < e2->x)
		return -1;
	else
		return 1;
}

static void
block_draw_envelope (Block *block, EnvelopeMode mode, GdkGC *gc)
{
	GtkWidget *widget = GTK_WIDGET (block);
	GList *list;
	gint x1, y1, x2, y2;
	gint x_scale;
	gfloat y_scale;
	gint width, height;
	Envelope *env;
	gboolean used_closest = TRUE;

        g_return_if_fail (block != NULL);
        g_return_if_fail (IS_BLOCK (widget));

	switch (mode) {
	case PAN_ENVELOPE:
		list = block->pan_env;
		break;
	case AMP_ENVELOPE:
		list = block->amp_env;
		break;
	default:
		g_assert_not_reached();
		return;
	};
	
	if (list == NULL)
		return;

	if (mode == block->envelope_mode)
		used_closest = FALSE;

	width = widget->allocation.width;
	height = widget->allocation.height;

	x_scale = block->track->song->x_scale;
	y_scale = BLOCK_HEIGHT / 2;


	env = (Envelope *)list->data;
	x1 = env->x / x_scale;
	y1 = env->y * y_scale;

	/* Draw first bullet */
	switch (mode) {
	case PAN_ENVELOPE:
		gdk_draw_arc (widget->window, gc, TRUE, x1 - RADIUS, height / 2 + y1 - RADIUS,
			      RADIUS * 2, RADIUS*2, 0, 360*64);
		break;
	case AMP_ENVELOPE:
		gdk_draw_arc (widget->window, gc, TRUE, x1 - RADIUS, height - y1 - RADIUS,
			      RADIUS * 2, RADIUS*2, 0, 360*64);
		break;
	default:
		g_assert_not_reached();
		return;
	};
	

	if (g_list_next (list)) {

		list = g_list_next (list);
		env = (Envelope *)list->data;
		x2 = env->x / x_scale;
		y2 = env->y * y_scale;
		
		if (used_closest == FALSE && block->closest_point && 
		    ((Envelope *)block->closest_point->data)->x / x_scale > x1 &&
		    ((Envelope *)block->closest_point->data)->x / x_scale < x2) {
			
			list = block->closest_point;
			env = (Envelope *)list->data;
			x2 = env->x / x_scale;
			y2 = env->y * y_scale;
			
			used_closest = TRUE;
		}
		
		while (list) {
			
			env = (Envelope *)list->data;
			x2 = env->x / x_scale;
			y2 = env->y * y_scale;
			
			if (used_closest == FALSE && block->closest_point && 
			    ((Envelope *)block->closest_point->data)->x / x_scale > x1 &&
			    ((Envelope *)block->closest_point->data)->x / x_scale < x2) {
				
				list = block->closest_point;
				env = (Envelope *)list->data;
				x2 = env->x / x_scale;
				y2 = env->y * y_scale;
				
				used_closest = TRUE;
			}
			
			while (g_list_next (list) && x2 > (((Envelope *)g_list_next(list)->data)->x / x_scale)) {
				
				list = g_list_next (list);
				env = (Envelope *)list->data;
				x2 = env->x / x_scale;
				y2 = env->y * y_scale;
				
				used_closest = TRUE;
			}
			
			switch (mode) {
			case PAN_ENVELOPE:
				gdk_draw_line (widget->window,
					       gc,
					       x1, height / 2 + y1,
					       x2, height / 2 + y2);
				break;
			case AMP_ENVELOPE:
				gdk_draw_line (widget->window,
					       gc,
					       x1, height - y1,
					       x2, height - y2);
				break;
			default:
				g_assert_not_reached();
				return;
			};
	
			x1 = x2;
			y1 = y2;
			
			switch (mode) {
			case PAN_ENVELOPE:
				gdk_draw_arc (widget->window, gc, TRUE, x1 - RADIUS, height / 2 + y1 - RADIUS,
					      RADIUS * 2, RADIUS*2, 0, 360*64);
				break;
			case AMP_ENVELOPE:
				gdk_draw_arc (widget->window, gc, TRUE, x1 - RADIUS, height - y1 - RADIUS,
					      RADIUS * 2, RADIUS*2, 0, 360*64);
				break;
			default:
				g_assert_not_reached();
				return;
			};
			list = g_list_next (list);
		};
	
	}
}

static gint
block_expose (GtkWidget      *widget,
	      GdkEventExpose *event)
{
        Block *block = BLOCK (widget);
	gint width, height;

        g_return_val_if_fail (widget != NULL, FALSE);
        g_return_val_if_fail (IS_BLOCK (widget), FALSE);
        g_return_val_if_fail (event != NULL, FALSE);

        if (event->count > 0)
		return FALSE;

	if (GTK_WIDGET_REALIZED (widget) == FALSE)
		return FALSE;

	width = widget->allocation.width;
	height = widget->allocation.height;

	if (block->pixmap == NULL)
		block_configure (widget, NULL);

	gdk_draw_pixmap(widget->window,
			block->wav_gc,
			block->pixmap,
			event->area.x, event->area.y,
			event->area.x, event->area.y,
			event->area.width,
			event->area.height);

	if (block->selection_flag) {
		gint selection_start = MIN(block->selection_start, block->selection_end) / block->track->song->x_scale;
		gint selection_width = ABS(block->selection_start - block->selection_end) / block->track->song->x_scale;

		/* First draw the wave pixmap */
		gdk_draw_pixmap(widget->window,
				block->wav_gc,
				block->pixmap,
				selection_start, event->area.y,
				selection_start, event->area.y,
				selection_width,
				event->area.height);
		/* Then invert it */
		gdk_draw_rectangle (widget->window,
				    block->selection_gc,
				    TRUE, selection_start, event->area.y,
				    selection_width, event->area.height);
	}
	block_draw_envelope (block, PAN_ENVELOPE, block->pan_gc);
	block_draw_envelope (block, AMP_ENVELOPE, block->amp_gc);

	return TRUE;
}

static gint
block_button_press (GtkWidget *widget,
			GdkEventButton *event)
{
        Block *block = BLOCK (widget);
	Envelope *env;
	gint x, tmpx, tmpy;
	gfloat y, distance;
	GList *list;
	GdkModifierType mods;

        g_return_val_if_fail (widget != NULL, FALSE);
        g_return_val_if_fail (IS_BLOCK (widget), FALSE);
        g_return_val_if_fail (event != NULL, FALSE);

	/* Get the Modifiers */
	gdk_window_get_pointer (widget->window, &tmpx, &tmpy, &mods);

	if (event->button == 1) {

	/* 
	 * If the user holds down the CTRL key when he press the left mouse button with a selection.
	 * That selection is extended.
	 */
		if (mods & GDK_CONTROL_MASK && block->selection_flag == TRUE) {
			gint x_pos = event->x * block->track->song->x_scale;
			
			if (ABS (x_pos - block->selection_start) > ABS (x_pos - block->selection_end))
				block->selection_end = x_pos;
			else {
				block->selection_start = block->selection_end;
				block->selection_end = x_pos;
			}
			/* Repaint the widget */
			gtk_widget_draw (widget, NULL);
		} else {
			/* Don't clame to have the selection any more */
			gtk_selection_owner_set (NULL, GDK_SELECTION_PRIMARY,
						 GDK_CURRENT_TIME);
			
			/* Set the start of the selection */
			block->selection_start = block->selection_end = event->x * block->track->song->x_scale;
		}

		gtk_grab_add (widget);
	
		x = event->x * block->track->song->x_scale;
		if (x < 0)
			x = 0;
		if (x >= block->length)
			x = block->length - 1;
		
		/* Look if we are near any envelope */
		block->envelope_mode = NO_ENVELOPE;
		list = NULL;
		block->closest_point = NULL;

		if (block->pan_env) {

			y = -1 + event->y * 2 / (gfloat)BLOCK_HEIGHT;
			distance = ABS (y - mixer_calc_envelope (block->pan_env, x));
			if (distance < MIN_Y_DISTANCE) {
				
				block->envelope_mode = PAN_ENVELOPE;
				list = block->pan_env;
			}
		}
		if (block->amp_env) {

			y = 2 - event->y * 2 / (gfloat)BLOCK_HEIGHT;
			distance = ABS (y - mixer_calc_envelope (block->amp_env, x));
			if (distance < MIN_Y_DISTANCE) {
				
				block->envelope_mode = AMP_ENVELOPE;
				list = block->amp_env;
			}
		}
		if (list) {

			distance = 1000000000;

			while (list) {
				env = (Envelope *)list->data;
				
				if (ABS (x - env->x) < distance) {
					
					distance = ABS (x - env->x);
					block->closest_point = list;
				}
				list = g_list_next (list);
			};
			distance /= block->track->song->x_scale;
			
			if (distance > MIN_DISTANCE || block->closest_point == NULL) {
				
				if (block->closest_point == NULL) {
					env = g_new0 (Envelope, 1);
					env->x = 0;
					
					switch (block->envelope_mode) {
					case PAN_ENVELOPE:
						env->y = 0;
						block->pan_env = g_list_append (block->pan_env, env);
						break;
					case AMP_ENVELOPE:
						env->y = 1;
						block->amp_env = g_list_append (block->amp_env, env);
						break;
					default:
						g_assert_not_reached();
						break;
					};
					env = g_new0 (Envelope, 1);
					env->x = block->length - 1;
					
					switch (block->envelope_mode) {
					case PAN_ENVELOPE:
						env->y = 0;
						block->pan_env = g_list_append (block->pan_env, env);
						break;
					case AMP_ENVELOPE:
						env->y = 1;
						block->amp_env = g_list_append (block->amp_env, env);
						break;
					default:
						g_assert_not_reached();
						break;
					};
				}
				
				env = g_new0 (Envelope, 1);
				env->x = event->x * block->track->song->x_scale;
				
				switch (block->envelope_mode) {
				case PAN_ENVELOPE:
					env->y = -1 + event->y * 2 / (gfloat)BLOCK_HEIGHT;

					block->pan_env = g_list_insert_sorted (block->pan_env, env, 
									       envelope_compare_func);
					block->closest_point = g_list_find (block->pan_env, env);
					break;
				case AMP_ENVELOPE:
					env->y = 2 - event->y * 2 / (gfloat)BLOCK_HEIGHT;

					block->amp_env = g_list_insert_sorted (block->amp_env, env, 
									       envelope_compare_func);
					block->closest_point = g_list_find (block->amp_env, env);	
					break;
				default:
					g_assert_not_reached();
					break;
				};
			
			gtk_widget_draw (widget, NULL);

			return TRUE;
			}
		}

		block->click_point.x = event->x;
		block->click_point.y = event->y;
		block->click_flag = TRUE;

	} else if (event->button == 3) {

		gtk_menu_popup (GTK_MENU (block->popup_menu), NULL, NULL, NULL, NULL,
				event->button, event->time);

	}
	return TRUE;
}

static gint
block_button_release (GtkWidget *widget,
		      GdkEventButton *event)
{
        Block *block = BLOCK (widget);
	GList *list;
        GdkModifierType mods;
	gint x, y;

        g_return_val_if_fail (widget != NULL, FALSE);
        g_return_val_if_fail (IS_BLOCK (widget), FALSE);
        g_return_val_if_fail (event != NULL, FALSE);

	block->click_flag = FALSE;

	if (event->window != widget->window)
		gdk_window_get_pointer (widget->window, &x, &y, &mods);

	if (event->button == 1) {

		block->copy_in_progress = FALSE;

		gtk_grab_remove (widget);

		if (block->envelope_mode != NO_ENVELOPE  && block->closest_point) {

			list = block->closest_point;

			if (g_list_next (list) && 
			    ((Envelope *)block->closest_point->data)->x > ((Envelope *)g_list_next(list)->data)->x) {

				g_free (block->closest_point->data);

				switch (block->envelope_mode) {
				case PAN_ENVELOPE:
					block->pan_env = g_list_remove (block->pan_env, block->closest_point->data); 
					break;
				case AMP_ENVELOPE:
					block->amp_env = g_list_remove (block->amp_env, block->closest_point->data); 
					break;
				default:
					g_assert_not_reached();
					break;
				};
			} else {
				while (g_list_previous (list) && 
				       ((Envelope *)block->closest_point->data)->x < ((Envelope *)g_list_previous (list)->data)->x) {
					switch (block->envelope_mode) {
					case PAN_ENVELOPE:
						block->pan_env = g_list_remove (block->pan_env, g_list_previous (list)->data); 
						break;
					case AMP_ENVELOPE:
						block->amp_env = g_list_remove (block->amp_env, g_list_previous (list)->data); 
						break;
					default:
						g_assert_not_reached();
						break;
					};
				}
			}
			block->closest_point = NULL;
		}

		/* Only clame the selection if it is more then 0 pixels wide */
		if (ABS(block->selection_end - block->selection_start) > 0) {
			gtk_selection_owner_set (widget,
						 GDK_SELECTION_PRIMARY,
						 GDK_CURRENT_TIME);

		} else if (event->y < widget->allocation.height / 2)
			song_move_marker (block->track->song, block->position + event->x * block->track->song->x_scale);
	}
	/* This vill take care of the overlapping of blocks */
	track_sort_blocks (block->track);

	if (event->button == 2)
		return TRUE;

	if (ABS(block->selection_end - block->selection_start) > 0)
		return TRUE;

	return FALSE;
}

gint 
block_selection_clear (GtkWidget *widget, 
		       GdkEventSelection *event)
{
        Block *block = BLOCK (widget);

	block->selection_flag = FALSE;
	block->selection_start = block->selection_end;

	if (block->track->song->selected_block == block)
		block->track->song->selected_block = NULL; /* I'm not selected anymore */

	gtk_widget_draw (widget, NULL);

	return TRUE;
}

static void
block_selection_get (GtkWidget *widget, 
		     GtkSelectionData *selection_data,
		     guint info,
		     guint time_stamp)
{
	Block *b;
	Block *block = BLOCK (widget);

	b = (Block *)block_new (block->track);

	b->start = block->start + MIN (block->selection_start, block->selection_end);
	b->length = ABS (block->selection_start - block->selection_end);
	b->sample = sample_ref (block->sample);
	b->position = block->position + MIN (block->selection_start, block->selection_end); 

        gtk_selection_data_set (selection_data, block_atom,
                                8, (gchar *)&b, sizeof (Block *));
}

static void
lower_left_box_motion (Block *block, GdkEventMotion *event, gint x, gint y)
{
	gint change, snap_pos;

	if (!block->click_flag)
		return;

	change = (block->click_point.x - x) * block->track->song->x_scale;

	if (prefs->snap_to_block) {
	
		snap_pos = song_get_snap_position (block->track->song, block, SNAP_POSITION);
		
		if (snap_pos > -1 && ABS (snap_pos - block->position) < block->track->song->x_scale && 
		    ABS (change) < SNAP_RESISTANCE * block->track->song->x_scale)
			change = 0;
	}

	/* Only valid values */
	
	/* Don't make the block to short */
	if (block->length + change < BLOCK_MIN_WIDTH)
		change = BLOCK_MIN_WIDTH - block->length;
	
	if (change > block->position)
		change = block->position;
	
	if (-change > block->length)
		change = block->length;
	
	if (change > block->start)
		change = block->start;
	
	block->position -= change;
	block->start -= change;
	block->length += change;

	/* Update the widget */
	block_update (block);
	/* The song need saving now */
	song_set_dirty (block->track->song, TRUE);
}

static void
lower_right_box_motion (Block *block, GdkEventMotion *event, gint x, gint y)
{
	gint change, snap_pos;

	if (!block->click_flag)
		return;

	change = (x - block->click_point.x) * block->track->song->x_scale;

	if (prefs->snap_to_block) {
		
		snap_pos = song_get_snap_position (block->track->song, block, SNAP_LENGTH);
		
		if (snap_pos > -1 && ABS(snap_pos - block->position) < block->track->song->x_scale && 
		    ABS(change) < SNAP_RESISTANCE * block->track->song->x_scale)
			change = 0;
	}

	/* Don't make the block to short */
	if (block->length + change < BLOCK_MIN_WIDTH)
		change = BLOCK_MIN_WIDTH - block->length;

	/* Don't make the block longer than the sample */
	if (change + block->start + block->length > 
	    block->sample->length / block->sample->byte_p_spl) {
		
		change = ((block->sample->length / block->sample->byte_p_spl)
			  - block->start - block->length);
	}
	block->length += change;
	block->click_point.x += change / block->track->song->x_scale;

	/* Update the widget */
	block_update (block);
	/* The song need saving now */
	song_set_dirty (block->track->song, TRUE);
}

static void
lower_half_motion (Block *block, GdkEventMotion *event, gint x, gint y)
{
	gint change, height;
	GtkWidget *widget = GTK_WIDGET (block);
	gint new_pos, snap_pos;

	height = widget->allocation.height;

	if (!block->click_flag)
		return;

	change = block->click_point.x - x;
	
	new_pos = MAX(block->position - block->track->song->x_scale * change, 0);

	if (prefs->snap_to_block) {
		
		snap_pos = song_get_snap_position (block->track->song, block, SNAP_POSITION | SNAP_LENGTH);
		
		if (snap_pos > -1 && ABS (snap_pos - block->position) < block->track->song->x_scale && 
		    ABS (change) < SNAP_RESISTANCE)
			new_pos = snap_pos;
	}
	block->position = new_pos;

	gdk_window_raise (widget->window);
	
	/* Change track? */
	if (y > height) {
		song_move_block(block->track->song, block, block->track->track_num + 1);
	}
	if (y < 0 && block->track->track_num) {
		song_move_block(block->track->song, block, block->track->track_num - 1);
	}

	/* Update the widget */
	block_update (block);
	/* The song need saving now */
	song_set_dirty (block->track->song, TRUE);
}

static void
upper_half_motion (Block *block, GdkEventMotion *event, gint x, gint y)
{
	g_assert (IS_BLOCK(block));

	if (!block->click_flag)
		return;

	block->track->song->selected_block = block;

	block->selection_flag = TRUE;
	block->selection_end = x * block->track->song->x_scale;

	/* check the boundaries */
	if (block->selection_end < 0)
		block->selection_end = 0;
	
	if (block->selection_end > block->length)
		block->selection_end = block->length;
	
	gtk_widget_draw (GTK_WIDGET(block), NULL);

	/* Update the widget */
	block_update (block);
}

static gint
block_motion_notify (GtkWidget *widget,
		     GdkEventMotion *event)
{
        Block *block = BLOCK (widget);
        GdkModifierType mods;
        gint x, y1, width, height;
	gfloat y;
	GdkCursorType new_type = GDK_TOP_LEFT_ARROW;
	gfloat distance;
	GList *list;
	Envelope *env;

        g_return_val_if_fail (widget != NULL, FALSE);
        g_return_val_if_fail (IS_BLOCK (widget), FALSE);
        g_return_val_if_fail (event != NULL, FALSE);

	if (event->is_hint || (event->window != widget->window))
		gdk_window_get_pointer (widget->window, &x, &y1, &mods);

	/* Change cursor if needed */
	x = event->x * block->track->song->x_scale;

	/* Make a + cursor if we are over an envelope */
	if (block->pan_env) {
		y = -1 + event->y * 2 / (gfloat)BLOCK_HEIGHT;

		distance = ABS (y - mixer_calc_envelope (block->pan_env, x));
		if (distance < MIN_Y_DISTANCE) {
			new_type = GDK_PLUS;
		}
	}
	if (block->amp_env) {
		y = 2 - event->y * 2 / (gfloat)BLOCK_HEIGHT;

		distance = ABS (y - mixer_calc_envelope (block->amp_env, x));
		if (distance < MIN_Y_DISTANCE) {
			new_type = GDK_PLUS;
		}
	}
	/* Make a Fleur cursor if we are over an envelope node */
	y = 2 - event->y * 2 / (gfloat)BLOCK_HEIGHT;
	list = block->pan_env;
	while (list) {
		env = (Envelope *)list->data;

		if (ABS (x - env->x) < MIN_DISTANCE * block->track->song->x_scale &&
		    ABS (y - env->y) < MIN_Y_DISTANCE) {

			new_type = GDK_FLEUR;
			break;
		}
		list = g_list_next (list);
	}
	y = -1 + event->y * 2 / (gfloat)BLOCK_HEIGHT;
	list = block->amp_env;
	while (list) {
		env = (Envelope *)list->data;

		if (ABS (x - env->x) < MIN_DISTANCE * block->track->song->x_scale &&
		    ABS (y - env->y) < MIN_Y_DISTANCE) {

			new_type = GDK_FLEUR;
			break;
		}
		list = g_list_next (list);
	}

	if (block->closest_point) {
		
		/* Don't change the x coordinate on the special nodes */
		if (((Envelope *)block->closest_point->data)->x != 0 &&
		    ((Envelope *)block->closest_point->data)->x != block->length - 1) {
			x = event->x * block->track->song->x_scale;
			if (x <= 0)
				x = block->track->song->x_scale;
			if (x >= block->length - 1)
				x = block->length - 2;
			
			((Envelope *)block->closest_point->data)->x = x;
		}
		/* Keep the y coordinates within the limits */
		switch (block->envelope_mode) {
		case PAN_ENVELOPE:
			y = -1 + event->y * 2 / (gfloat)BLOCK_HEIGHT;
			if (y < -1)
				y = -1;
			if (y > 1)
				y = 1;
			
			break;
		case AMP_ENVELOPE:
			y = 2 - event->y * 2 / (gfloat)BLOCK_HEIGHT;
			if (y < 0)
				y = 0;
			if (y > 2)
				y = 2;			
			break;
		default:
			g_assert_not_reached();
			break;
		};
		((Envelope *)block->closest_point->data)->y = y;
		
		gtk_widget_draw (widget, NULL);

		/* The song need saving now */
		song_set_dirty (block->track->song, TRUE);
		
	} else {

		if (event->is_hint || (event->window != widget->window))
			gdk_window_get_pointer (widget->window, &x, &y1, &mods);

		width = widget->allocation.width;
		height = widget->allocation.height;

		/* Draging lower left box */
		if ( block->click_point.x <= BLOCK_HANDLE_SIZE && 
		     block->click_point.y >= height - BLOCK_HANDLE_SIZE) {
			
			gtk_signal_emit (GTK_OBJECT (block), 
					 block_signals[LOWER_LEFT_BOX_MOTION], event, x, y1);
			
			
			/* Dragin lower right box */
		} else if ( block->click_point.x >= width - BLOCK_HANDLE_SIZE &&
			    block->click_point.y >= height - BLOCK_HANDLE_SIZE) {
			
			gtk_signal_emit (GTK_OBJECT (block), 
					 block_signals[LOWER_RIGHT_BOX_MOTION], event, x, y1);
			
		} else if (block->click_point.y < height / 2) {
				/* Only move and stretch on the lower half of the block */
			
			gtk_signal_emit (GTK_OBJECT (block), 
					 block_signals[UPPER_HALF_MOTION], event, x, y1);
			
		} else { /* Move */
			
			/*
			  If the user press shift while moving the block, we'll leave a copy of it and 
			  move the original block.
			*/
			if (mods & GDK_SHIFT_MASK && block->copy_in_progress == FALSE) {
				/* Copy the block before moving.... */
				
				Block *newBlock = BLOCK (block_new (block->track));
				newBlock->sample = sample_ref (block->sample);
				newBlock->start = block->start;
				newBlock->length = block->length;
				
				/* Position it on the same place as the popup menu */
				newBlock->position = block->position;
				
				/* Insert the newBlock into the song */
				song_insert_block (block->track->song, newBlock, block->track->track_num);
				
				
				block->copy_in_progress = TRUE;
				/* OK the block is copied ....
				   can move the old one....*/
				
			}
			gtk_signal_emit (GTK_OBJECT (block), 
					 block_signals[LOWER_HALF_MOTION], event, x, y1);
			
		}
	}		
	/* Change the cursor if needed */
	if (new_type != (GdkCursorType) block->cursor_type) {
		GdkCursor *cursor;
                      
		block->cursor_type = new_type;

		cursor = gdk_cursor_new (block->cursor_type);
		gdk_window_set_cursor (widget->window, cursor);
		gdk_cursor_destroy (cursor);
        }

        return TRUE;
}

/*
 * Function draws the entire widget, including waveform and envelopes
 * really just a big hack
 */
void
block_draw (Block *block)
{
	gint i, width, height;
	gint offset, value;
	GString * start_gs , * len_gs ;

	g_assert (IS_BLOCK(block));

	width = GTK_WIDGET(block)->allocation.width;
	height = GTK_WIDGET(block)->allocation.height;

	if ((block->old_width != width || block->old_height != height) && block->pixmap) {
		gdk_pixmap_unref(block->pixmap);
		block->pixmap = NULL;
	}

	if (block->pixmap == NULL)
		block->pixmap = gdk_pixmap_new (GTK_WIDGET(block)->window, width, height, -1);

	block->old_width = width;
	block->old_height = height;

	gdk_draw_rectangle (block->pixmap,
			    GTK_WIDGET(block)->style->white_gc,
			    TRUE,
			    0,
			    0,
			    width,
			    height);

	offset = block->start / INDEX_RATIO;

	for (i = 0; i < width;i++) {
		value = index_get_average (block->sample->index, 
					   offset, block->start / INDEX_RATIO,
					   block->track->song->x_scale / INDEX_RATIO, 
					   block->sample->length / INDEX_RATIO / 2) / ((float)0x7f / height);
		gdk_draw_line (block->pixmap,
			       block->wav_gc,
			       i, height / 2 + value,
			       i, height / 2 - value);
		
		offset += block->track->song->x_scale / INDEX_RATIO;
	}
	
	
	gdk_draw_rectangle (block->pixmap,
			    GTK_WIDGET(block)->style->fg_gc[GTK_WIDGET(block)->state],
			    FALSE,
			    0,
			    0,
			    width - 1,
			    height - 1);

	/* Draw a black "handle" in the lower left corner */
        gdk_draw_rectangle (block->pixmap,
                            GTK_WIDGET(block)->style->black_gc,
                            TRUE,
                            0,
                            height - BLOCK_HANDLE_SIZE,
                            BLOCK_HANDLE_SIZE,
                            BLOCK_HANDLE_SIZE);

	/* Draw a black "handle" in the lower right corner */
        gdk_draw_rectangle (block->pixmap,
                            GTK_WIDGET(block)->style->black_gc,
                            TRUE,
                            width - BLOCK_HANDLE_SIZE,
                            height - BLOCK_HANDLE_SIZE,
                            BLOCK_HANDLE_SIZE,
                            BLOCK_HANDLE_SIZE);

	/* Draw title */

	gdk_draw_text (block->pixmap,
		       GTK_WIDGET(block)->style->font,
		       GTK_WIDGET(block)->style->fg_gc[GTK_WIDGET(block)->state],
		       0, 10,
		       block->sample->id,
		       strlen (block->sample->id));


	/* Draw start time */

	start_gs = g_string_sized_new(20);

	start_gs->len = g_snprintf(start_gs->str,
				   20,
				   "%02.0f:%02.0f.%02.0f",
				 floor(block->position / (60*SAMPLE_RATE)),
				 (block->position / (SAMPLE_RATE)) - 
				    60*floor(block->position / (60*SAMPLE_RATE)),
				 (block->position / (SAMPLE_RATE/100)) -
				    100*floor(block->position / (SAMPLE_RATE)));




	gdk_draw_text  (block->pixmap,
		       GTK_WIDGET(block)->style->font,
		       GTK_WIDGET(block)->style->fg_gc[GTK_WIDGET(block)->state],
		       0, height - 20,
		       "S:",
		       2);

	gdk_draw_text  (block->pixmap,
		       GTK_WIDGET(block)->style->font,
		       GTK_WIDGET(block)->style->fg_gc[GTK_WIDGET(block)->state],
		       15, height - 20,
		       start_gs->str,
		       start_gs->len);


	/* Draw length */


	len_gs = g_string_sized_new(10);

	len_gs->len = g_snprintf(len_gs->str,
				   20,
				   "%02.0f:%02.0f.%02.0f",
				 floor(block->length / (60*SAMPLE_RATE)),
				 (block->length / (SAMPLE_RATE)) - 
				    60*floor(block->length / (60*SAMPLE_RATE)),
				 (block->length / (SAMPLE_RATE/100)) -
				    100*floor(block->length / (SAMPLE_RATE)));




	gdk_draw_text (block->pixmap,
		       GTK_WIDGET(block)->style->font,
		       GTK_WIDGET(block)->style->fg_gc[GTK_WIDGET(block)->state],
		       0, height - 5,
		       "L:",
		       2);

	gdk_draw_text (block->pixmap,
		       GTK_WIDGET(block)->style->font,
		       GTK_WIDGET(block)->style->fg_gc[GTK_WIDGET(block)->state],
		       15, height - 5,
		       len_gs->str,
		       len_gs->len);



	block_draw_envelope (block, PAN_ENVELOPE, block->pan_gc);
	block_draw_envelope (block, AMP_ENVELOPE, block->amp_gc);	
}

static gboolean 
block_configure (GtkWidget *widget,GdkEventConfigure *event)
{
	gint width, height;
	Block *block = BLOCK (widget);

	width = widget->allocation.width;
	height = widget->allocation.height;

	if (block->wav_gc == NULL)
		return TRUE;

	block_draw(block);

	return TRUE;
}

static void
block_check_envelopes (Block *block)
{
	Envelope *env;
	GList *list, *last;
	gint max_x;

	if (block->amp_env) {

		last = g_list_last (block->amp_env);
		env = (Envelope *)last->data;

		max_x = env->x = block->length - 1;

		list = g_list_previous (last);
		while (list) {
			env = (Envelope *)list->data;

			if (env->x > max_x) {
				block->amp_env = g_list_remove (block->amp_env, env);
				list = g_list_previous (last);

			} else
				break;
		}
	}
	if (block->pan_env) {

		last = g_list_last (block->pan_env);
		env = (Envelope *)last->data;

		max_x = env->x = block->length - 1;

		list = g_list_previous (last);
		while (list) {
			env = (Envelope *)list->data;

			if (env->x > max_x) {
				block->pan_env = g_list_remove (block->pan_env, env);
				list = g_list_previous (last);
			} else
				break;
		}
	}
}

void
block_update (Block *block)
{
	block_check_envelopes (block);

	gtk_layout_move (GTK_LAYOUT (GTK_WIDGET (block)->parent), GTK_WIDGET (block), 
			 block->position / block->track->song->x_scale, 0);

	gtk_widget_set_usize (GTK_WIDGET (block), 
			      block->length / block->track->song->x_scale,
			      block->track->song->block_height);

	song_update_scrollbar (block->track->song);
}

/*
 * User clicked 'delete'.
 */
static void
block_delete_clicked (GtkWidget *widget, Block *block)
{
	song_delete_block (block->track->song, block);
}

static void
block_add_amp_envelope (GtkWidget *widget, Block *block)
{
	Envelope *env;

	if (block->amp_env == NULL) {

		env = g_new0 (Envelope, 1);
		env->x = 0;
		env->y = 1;
		block->amp_env = g_list_append (block->amp_env, env);

		env = g_new0 (Envelope, 1);
		env->x = block->length - 1;
		env->y = 1;
		block->amp_env = g_list_append (block->amp_env, env);

		gtk_widget_draw (GTK_WIDGET (block), NULL);
	}
}

static void
block_add_pan_envelope (GtkWidget *widget, Block *block)
{
	Envelope *env;

	if (block->pan_env == NULL) {

		env = g_new0 (Envelope, 1);
		env->x = 0;
		env->y = 0;
		block->pan_env = g_list_append (block->pan_env, env);

		env = g_new0 (Envelope, 1);
		env->x = block->length - 1;
		env->y = 0;
		block->pan_env = g_list_append (block->pan_env, env);

		gtk_widget_draw (GTK_WIDGET (block), NULL);
	}
}

static void
block_remove_amp_envelope (GtkWidget *widget, Block *block)
{
	/* Free amplitude envelope memory */
	g_list_foreach (block->amp_env, free_envelope, NULL);
	g_list_free (block->amp_env);
	block->amp_env = NULL;

	gtk_widget_draw (GTK_WIDGET (block), NULL);
}

static void
block_remove_pan_envelope (GtkWidget *widget, Block *block)
{
	/* Free pan envelope memory */
	g_list_foreach (block->pan_env, free_envelope, NULL);
	g_list_free (block->pan_env);
	block->pan_env = NULL;

	gtk_widget_draw (GTK_WIDGET (block), NULL);
}

static void
block_select_all (GtkWidget *widget, Block *block)
{
	/* Select the whole block */
	block->selection_start = 0;
	block->selection_end = block->length;
	block->selection_flag = TRUE;
	gtk_selection_owner_set (GTK_WIDGET (block),
				 GDK_SELECTION_PRIMARY,
				 GDK_CURRENT_TIME);
	/* Redraw the widget */
	gtk_widget_draw (GTK_WIDGET (block), NULL);
}

static void
block_convert_to_loopblock (GtkWidget *widget, Block *block)
{
	LoopBlock *lblock;

	lblock = LOOPBLOCK (loopblock_new_from_block (block, block->track));
	
	/* Insert the loopblock into the song */
	song_insert_block (block->track->song, BLOCK (lblock), block->track->track_num);

	/* Remove the old block */
	song_delete_block (block->track->song, block);
}

static GnomeUIInfo add_submenu[] = {
	GNOMEUIINFO_ITEM_NONE (N_("Amplitude envelope"), N_("Add amplitude envelope"), block_add_amp_envelope),
	GNOMEUIINFO_ITEM_NONE (N_("Pan envelope"), N_("Add pan envelope"), block_add_pan_envelope),
	GNOMEUIINFO_END
};
static GnomeUIInfo delete_submenu[] = {
	GNOMEUIINFO_ITEM_NONE (N_("Amplitude envelope"), N_("Delete amplitude envelope"), block_remove_amp_envelope),
	GNOMEUIINFO_ITEM_NONE (N_("Pan envelope"), N_("Delete pan envelope"), block_remove_pan_envelope),
	GNOMEUIINFO_SEPARATOR,
	GNOMEUIINFO_ITEM_NONE (N_("Block"), N_("Delete Block"), block_delete_clicked),
	GNOMEUIINFO_END
};



static GnomeUIInfo block_menu[] = {
	GNOMEUIINFO_SUBTREE (N_("Add"), add_submenu),
	GNOMEUIINFO_SUBTREE (N_("Delete"), delete_submenu),
	GNOMEUIINFO_END
};

static GnomeUIInfo block_specific_menu[] = {
	GNOMEUIINFO_SEPARATOR,
	GNOMEUIINFO_ITEM_NONE (N_("Convert to loopblock"), N_("Convert to loopblock"), block_convert_to_loopblock),
	GNOMEUIINFO_SEPARATOR,
	GNOMEUIINFO_ITEM_NONE (N_("Select all"), N_("Make the selection cover the whole block"), block_select_all),
	GNOMEUIINFO_END
};

static void 
block_tool_cb(GtkWidget *widget, PluginCBInfo *info)
{
#define tool ((ToolInfo *)info->plugin_info)
#define block ((Block *)info->local)
	g_assert(info != NULL);
	g_assert(info->plugin_info != NULL);
	g_assert(info->local != NULL);

	if (tool->execute == NULL)
		g_warning("Tool cannot execute");
	else
		tool->execute(tool, block->track->song, block);
#undef block
#undef tool
}

static void
block_create_popup_menu (Block *block)
{

	/* Create a popup menu */
        block->popup_menu = gtk_menu_new();

	gnome_app_fill_menu_with_data (GTK_MENU_SHELL (block->popup_menu), 
		block_menu, NULL, FALSE, 0, block);

}

/*
 * This function frees the PluginCBInfo-structure
 */
static void
block_tool_menu_item_destroy (GtkWidget *widget, PluginCBInfo *toolinfo) 
{
	g_assert(toolinfo != NULL);

	g_free(toolinfo);
}

GtkWidget*
block_new (Track *track)
{
	Block *block;
	GList *node;
	GtkWidget *menu;
	GtkWidget *item;
	PluginCBInfo *cbinfo;

        block = BLOCK (gtk_type_new (block_get_type ()));

	/* Add block specific items to the popup menu */
	gnome_app_fill_menu_with_data (GTK_MENU_SHELL (block->popup_menu), 
		block_specific_menu, NULL, FALSE, (sizeof (block_menu) / sizeof(GnomeUIInfo)) - 1, block);

	
	/* Create the Tools-menu */
	item = gtk_menu_item_new_with_label(_("Tools"));
	gtk_menu_append(GTK_MENU(block->popup_menu), item);
	gtk_widget_show(item);

	menu = gtk_menu_new();
	gtk_menu_item_set_submenu(GTK_MENU_ITEM(item), menu);

	/* Add all the tools */
	for (node = plugins_get_tool_list();node;node = node->next) {
		g_assert(node->data != NULL);

		cbinfo = g_new0(PluginCBInfo, 1); 

		cbinfo->plugin_info = node->data;
		cbinfo->local = (gpointer)block;

		item = gtk_menu_item_new_with_label(((ToolInfo *)node->data)->name);
		gtk_menu_append(GTK_MENU(menu), item);

		gtk_signal_connect(GTK_OBJECT(item), "activate", GTK_SIGNAL_FUNC(block_tool_cb), cbinfo);
		gtk_signal_connect(GTK_OBJECT(item), "destroy", GTK_SIGNAL_FUNC(block_tool_menu_item_destroy), cbinfo);
	}

	gtk_widget_show_all(menu);

	block->track = track;

	return GTK_WIDGET (block);
}

void
block_mix_into_buffer (Block *block, gint16 *buffer, gint position, gint length) {
	gtk_signal_emit (GTK_OBJECT (block), 
			 block_signals[MIX_INTO_BUFFER_SIGNAL], 
			 buffer, position, length);
}

/* 
 * This function is called by the mixer.
 */

static void
mix_into_buffer (Block *block, gint16 *buffer, gint position, gint length)
{
	gint16 *data;
	gint i;
	
	data = (gint16 *)(block->sample->data + 
			  (block->start + position - block->position) * block->sample->byte_p_spl);
		
	for (i = 0;i < length / (sizeof (guint16) * 2);i++) {

		mixer_mix_sample_into_buffer (buffer,
					      data,
					      block->amp_env, 
					      block->pan_env, 
					      block->track->amp,
					      block->track->pan,
					      position - block->position + i, block->sample->num_channels);
		
		data += block->sample->num_channels;
		buffer += 2;
	}
}

/*
 * Loads a block, and inserts it into the specified track.
 * Returns TRUE if successfull
 */
gboolean
block_load (Song *song, Track *track, xmlNodePtr node, xmlNsPtr ns, gint track_num) {
	Block *block;
	gchar *ptr;
	xmlNodePtr cur;
	Envelope *envelope;

	block = BLOCK (block_new (track));
	
	block->start = atoi ((ptr = xmlGetProp (node, "start")));
	g_free (ptr);
	block->length = atoi ((ptr = xmlGetProp (node, "length")));
	g_free (ptr);
	block->position = atoi ((ptr = xmlGetProp (node, "position")));
	g_free (ptr);
	ptr = xmlGetProp (node, "sample");

	if ((block->sample = song_get_sample_by_id(song, ptr)) == NULL) {
		g_free (ptr);
		gtk_widget_destroy(GTK_WIDGET(block));
		return FALSE;
	}

	g_free (ptr);
	
	/* Add the block to out list */
	song_insert_block (song, block, track_num);
	
	cur = node->childs;
	
	while (cur) {
		
		if ((!strcmp(cur->name, "ampenv")) && (cur->ns == ns)) {
			envelope = g_new0 (Envelope, 1);
			
			envelope->x = atoi ((ptr = xmlGetProp (cur, "x")));
			g_free (ptr);
			envelope->y = atof ((ptr = xmlGetProp (cur, "y")));
			g_free (ptr);
			block->amp_env = g_list_append (block->amp_env, envelope);
		}
		
		if ((!strcmp(cur->name, "panenv")) && (cur->ns == ns)) {
			envelope = g_new0 (Envelope, 1);
			
			envelope->x = atoi ((ptr = xmlGetProp (cur, "x")));
			g_free (ptr);
			envelope->y = atof ((ptr = xmlGetProp (cur, "y")));
			g_free (ptr);
			block->pan_env = g_list_append (block->pan_env, envelope);
		}
		cur = cur->next;
	};

	return TRUE;
}
