/* -*- Mode: C; tab-width: 8; indent-tabs-mode: t; c-basic-offset: 8 -*- */
/*  This file is part of MixMagic.

    Copyright (C) 2000-2001 Jonas Borgstrm <jonas@bengburken.net>

    This program is free software; you can redistribute it and/or
    modify it under the terms of the GNU General Public
    License as published by the Free Software Foundation; either
    version 2 of the License, or (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the 
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software 
    Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
*/

#include "mixmagic.h"
#include "mixer.h"
#include "mainwin.h"
#include "track.h"
#include "block.h"
#include "sample.h"
#include "song.h"
#include "stock.h"
#include "preferences.h"
#include "sidebar.h"

static GList *window_list = NULL; /* This list contains pointers to all 
				     open Mainwin classes */

static GdkAtom block_atom = GDK_NONE;
static GtkObjectClass *parent_class = NULL;

void about_show (const GtkWidget *parent_win);
static void mainwin_destroy (GtkObject *object);
static void mainwin_class_init (MainwinClass *class);
static void mainwin_init (Mainwin *mainwin);
static void mainwin_import_clicked (GtkWidget *widget, gpointer data);
static void mainwin_zoom_in_clicked (GtkWidget *widget, Mainwin *mainwin);
static void mainwin_zoom_out_clicked (GtkWidget *widget, Mainwin *mainwin);
static void mainwin_zoom_fit_clicked (GtkWidget *widget, Mainwin *mainwin);
static void mainwin_song_info_clicked (GtkWidget *widget, Mainwin *mainwin);
static void mainwin_close_clicked (GtkWidget *widget, Mainwin *mainwin);
static void mainwin_exit_clicked (GtkWidget *widget, gpointer data);
static gboolean mainwin_delete_event (GtkWidget *widget, GdkEventAny *event, Mainwin *mainwin);
static void mainwin_new_clicked (GtkWidget *widget, gpointer data);
static void mainwin_preferences_clicked (GtkWidget *widget, gpointer data);
static gboolean mainwin_show_open_dlg (Mainwin *mainwin, gchar **filename);
static void mainwin_open_clicked (GtkWidget *widget, gpointer data);
static void mainwin_about_clicked (GtkWidget *widget, gpointer data);
static void mainwin_repeat_toggled (GtkWidget *widget, Mainwin *mainwin);
static void mainwin_selection_received (GtkWidget *widget, GtkSelectionData *selection_data, guint time, gpointer data);
static void mainwin_rewind_clicked (GtkWidget *widget, Mainwin *mainwin);
static gboolean mainwin_close_if_user_permits (Mainwin *mw);
static void mainwin_sidebar_show_cb (GtkWidget *widget, Mainwin *mainwin);
static void mainwin_sidebar_hide_cb (GtkWidget *widget, Mainwin *mainwin);
static void mainwin_sidebar_toggle_cb (GtkWidget *widget, Mainwin *mainwin);
static void mainwin_size_allocate (GtkWidget *widget, GtkAllocation *allocation, gpointer data);

/*
 * Mainwin menus 
 */

/* Don't forget to update "file_menu_index" if you change the menu structure */

static GnomeUIInfo file_menu [] = {
	GNOMEUIINFO_MENU_NEW_ITEM (N_("New song"), N_("Create a new song."), mainwin_new_clicked, NULL),
	GNOMEUIINFO_MENU_OPEN_ITEM (mainwin_open_clicked, NULL),
	GNOMEUIINFO_MENU_SAVE_ITEM (mainwin_save, NULL),
	GNOMEUIINFO_MENU_SAVE_AS_ITEM (mainwin_save_as, NULL),
	GNOMEUIINFO_SEPARATOR,
	GNOMEUIINFO_ITEM_NONE (N_("Import song"), N_("Import an existing song"), mainwin_import_clicked),
	GNOMEUIINFO_SEPARATOR,
	GNOMEUIINFO_MENU_CLOSE_ITEM (mainwin_close_clicked, NULL),
	GNOMEUIINFO_MENU_EXIT_ITEM (mainwin_exit_clicked, NULL),
	GNOMEUIINFO_END
};

#define SIDEBAR_INDEX 0
static GnomeUIInfo view_menu[] = {
	GNOMEUIINFO_TOGGLEITEM (N_("Show sidebar"), 
				N_("Show a sidebar on the left side of the screen."), 
				mainwin_sidebar_toggle_cb, NULL),
	GNOMEUIINFO_SEPARATOR,
	GNOMEUIINFO_ITEM_NONE (N_("Zoom in"), N_("Zoom in"), mainwin_zoom_in_clicked),
	GNOMEUIINFO_ITEM_NONE (N_("Zoom out"), N_("Zoom out"), mainwin_zoom_out_clicked),
	GNOMEUIINFO_ITEM_NONE (N_("Zoom fit"), N_("Zoom fit"), mainwin_zoom_fit_clicked),
	GNOMEUIINFO_SEPARATOR,
	GNOMEUIINFO_ITEM_NONE (N_("Song-Info"), N_("View song-information"), mainwin_song_info_clicked),
	GNOMEUIINFO_END
};

static GnomeUIInfo song_menu[] = {
	GNOMEUIINFO_ITEM_NONE (N_("Play"), N_("Play song"), mainwin_play_clicked),
	GNOMEUIINFO_ITEM_NONE (N_("Stop"), N_("Stop song"), mainwin_stop_clicked),
	GNOMEUIINFO_END
};

static GnomeUIInfo settings_menu[] = {
	GNOMEUIINFO_MENU_PREFERENCES_ITEM (mainwin_preferences_clicked, NULL),
	GNOMEUIINFO_END
};

static GnomeUIInfo help_menu [] = {
	GNOMEUIINFO_MENU_ABOUT_ITEM (mainwin_about_clicked, NULL),
	GNOMEUIINFO_END
};

static GnomeUIInfo main_menu[] = {
	GNOMEUIINFO_MENU_FILE_TREE (file_menu),
	GNOMEUIINFO_MENU_VIEW_TREE (view_menu),
	GNOMEUIINFO_SUBTREE (N_("_Song"), song_menu),
	GNOMEUIINFO_MENU_SETTINGS_TREE (settings_menu),
	GNOMEUIINFO_MENU_HELP_TREE (help_menu),
	GNOMEUIINFO_END
};

#define TOOLBAR_NEW_INDEX 0
#define TOOLBAR_OPEN_INDEX 1
#define TOOLBAR_SAVE_INDEX 2
#define TOOLBAR_REWIND_INDEX 4
#define TOOLBAR_STOP_INDEX 5
#define TOOLBAR_PLAY_INDEX 6
#define TOOLBAR_REPEAT_INDEX 7
#define TOOLBAR_ZOOMIN_INDEX 9
#define TOOLBAR_ZOOMOUT_INDEX 10 
#define TOOLBAR_ZOOMFIT_INDEX 11

static GnomeUIInfo mixmagic_toolbar[] = {
	GNOMEUIINFO_ITEM_STOCK (N_("New"), N_("Creates a new song."),
				mainwin_new_clicked, GNOME_STOCK_PIXMAP_NEW),
	GNOMEUIINFO_ITEM_STOCK (N_("Open"), N_("Open an existing song."),
				mainwin_open_clicked, GNOME_STOCK_PIXMAP_OPEN),
	GNOMEUIINFO_ITEM_STOCK (N_("Save"), N_("Save the song to a file."),
				mainwin_save, GNOME_STOCK_PIXMAP_SAVE),
	GNOMEUIINFO_SEPARATOR,
	GNOMEUIINFO_ITEM_STOCK (N_("Rewind"), N_("Rewind to the beginning."),
				mainwin_rewind_clicked, STOCK_REWIND),
	GNOMEUIINFO_ITEM_STOCK (N_("Stop"), N_("Stop the playback."),
				mainwin_stop_clicked, STOCK_STOP),
	GNOMEUIINFO_ITEM_STOCK (N_("Play"), N_("Start playback."),
				mainwin_play_clicked, STOCK_PLAY),
	{GNOME_APP_UI_TOGGLEITEM, N_("Repeat"), N_("Restart at the beginning, when the player gets to the end of the song."),
	 mainwin_repeat_toggled, NULL, NULL, GNOME_APP_PIXMAP_STOCK, GNOME_STOCK_PIXMAP_REFRESH },
	GNOMEUIINFO_SEPARATOR,
	GNOMEUIINFO_ITEM_STOCK (N_("Zoom in"), N_("Zoom in."),
				mainwin_zoom_in_clicked, STOCK_ZOOM_IN),
	GNOMEUIINFO_ITEM_STOCK (N_("Zoom out"), N_("Zoom out."),
				mainwin_zoom_out_clicked, STOCK_ZOOM_OUT),
	GNOMEUIINFO_ITEM_STOCK (N_("Zoom fit"), N_("Zoom fit."),
				mainwin_zoom_fit_clicked, STOCK_ZOOM_FIT),
	GNOMEUIINFO_END
};

guint
mainwin_get_type ()
{
        static guint mainwin_type = 0;
	
        if (!mainwin_type) {

		GtkTypeInfo mainwin_info = {
			"Mainwin",
			sizeof (Mainwin),
			sizeof (MainwinClass),
			(GtkClassInitFunc) mainwin_class_init,
			(GtkObjectInitFunc) mainwin_init,
			(GtkArgSetFunc) NULL,
			(GtkArgGetFunc) NULL
		};
	    
		mainwin_type = gtk_type_unique (gtk_object_get_type (), &mainwin_info);
          }
	
        return mainwin_type;
}

static void
mainwin_class_init (MainwinClass *class)
{
	GtkObjectClass *object_class;

	parent_class = gtk_type_class (GTK_TYPE_OBJECT);

	object_class = (GtkObjectClass*) class;
	object_class->destroy = mainwin_destroy;
}

static void
mainwin_init (Mainwin *mainwin)
{
}

static void 
mainwin_destroy (GtkObject *object)
{
	Mainwin *mainwin = MAINWIN (object);

	g_print ("mainwin_destroy\n");

	if (mainwin->filename)
		g_free (mainwin->filename);

	gtk_widget_destroy (mainwin->toplevel);

	if (mainwin->open_file_sel)
		gtk_widget_destroy (mainwin->open_file_sel);

	if (mainwin->save_file_sel)
		gtk_widget_destroy (mainwin->save_file_sel);

	(*GTK_OBJECT_CLASS (parent_class)->destroy) (object);
}

/*
 * Adds a toolbar to the mainwin 
 */
static void
mainwin_create_toolbars (Mainwin *mainwin)
{
	GtkAccelGroup *accel_group;

	gnome_app_create_toolbar_with_data (GNOME_APP (mainwin->toplevel), mixmagic_toolbar, mainwin);
	
	accel_group = gtk_accel_group_new ();
       	gtk_accel_group_attach (accel_group, GTK_OBJECT (mainwin->toplevel));
	
	gtk_widget_add_accelerator (mixmagic_toolbar[TOOLBAR_PLAY_INDEX].widget,
				    "clicked",
				    accel_group,
				    GDK_space,
				    0,
				    GTK_ACCEL_VISIBLE | GTK_ACCEL_SIGNAL_VISIBLE);

}

/*
 * Creates a window 
*/
GtkObject *
mainwin_new(const gchar *filename)
{
	Mainwin *mainwin;
	gchar *title;
	gboolean show_sidebar;
	gint default_width, default_height , ruler_pad , ruler_start , ruler_end , ruler_width;

	/* Allocate new structure */
	mainwin = MAINWIN (gtk_type_new (mainwin_get_type ()));

	window_list = g_list_append (window_list, mainwin);

	mainwin->toplevel = gnome_app_new (_("mixmagic"), _("MixMagic"));
	
	gtk_signal_connect (GTK_OBJECT (mainwin->toplevel), "delete_event",
			    GTK_SIGNAL_FUNC (mainwin_delete_event), mainwin);

	gtk_signal_connect (GTK_OBJECT (mainwin->toplevel), "size_allocate",
			    GTK_SIGNAL_FUNC (mainwin_size_allocate), mainwin);

	gtk_signal_connect (GTK_OBJECT (mainwin->toplevel), "selection_received",
			    GTK_SIGNAL_FUNC (mainwin_selection_received), mainwin);


	gnome_app_create_menus_with_data (GNOME_APP (mainwin->toplevel), main_menu, mainwin);

	mainwin_create_toolbars (mainwin);

	mainwin->appbar = gnome_appbar_new (TRUE, TRUE, GNOME_PREFERENCES_USER);
	gnome_app_set_statusbar (GNOME_APP (mainwin->toplevel), mainwin->appbar);

	/* Create a new song class */
	mainwin->song = SONG (song_new (GNOME_APPBAR (mainwin->appbar)));

	/* Load a file if a filename is specified */
	if (filename != NULL) {
		if (song_load (mainwin->song, filename)) {
			
			mainwin->filename = g_strdup (filename);
			title = g_strdup_printf (_("MixMagic: %s"), g_basename (filename));
			gtk_window_set_title (GTK_WINDOW (mainwin->toplevel), title);
			g_free (title);
		} else {
			gchar *error;
			
			error = g_strdup_printf (_("Failed to open '%s'.\nIs it a .mmsong file?"),
						 filename);
			gnome_error_dialog (error);
			g_free (error);			
		}
	} else { 
		/* Here should we disable everything that shouldn't be clickable when there is no
		* song loaded
		*
		* gtk_widget_set_sensitive (GTK_WIDGET (mixmagic_toolbar[TOOLBAR_SAVE_INDEX].widget), FALSE);
		*/
	}

	mainwin->sidebar = sidebar_new ();

	gtk_signal_connect (GTK_OBJECT(mainwin->sidebar), 
			   "close_clicked", 
			   GTK_SIGNAL_FUNC (mainwin_sidebar_hide_cb), mainwin);

	mainwin->scrollwin = gtk_scrolled_window_new ( GTK_ADJUSTMENT (gtk_adjustment_new ( 0, 0, 0, 0, 0, 0 )), 
						       GTK_ADJUSTMENT (gtk_adjustment_new ( 0, 0, 0, 0, 0, 0 )));
	gtk_scrolled_window_add_with_viewport ( GTK_SCROLLED_WINDOW (mainwin->scrollwin) , GTK_WIDGET (mainwin->song) );
	gtk_scrolled_window_set_policy ( GTK_SCROLLED_WINDOW (mainwin->scrollwin) , GTK_POLICY_NEVER , GTK_POLICY_AUTOMATIC );
	mainwin->songwin = gtk_vbox_new ( FALSE , 0 );

	mainwin->song->rulerwin = gtk_hbox_new ( FALSE , 0 );


	mainwin->song->ruler_label = gtk_button_new_with_label ( _("Time in Seconds") );
	gtk_widget_show ( mainwin->song->ruler_label ); 

		gtk_box_pack_start ( GTK_BOX ( mainwin->song->rulerwin ) , mainwin->song->ruler_label , FALSE , TRUE , 19 );

	mainwin->song->ruler = gtk_hruler_new();
	ruler_width = TRACK (mainwin->song->track_list->data)->layout->allocation.width;
	ruler_start = mainwin->song->marker_position / 44100 ;
	ruler_end   = mainwin->song->marker_position + ruler_width * mainwin->song->x_scale / 4100 ;
	gtk_ruler_set_range ( GTK_RULER ( mainwin->song->ruler ) , ruler_start , ruler_end , 0 , ruler_width );
	gtk_widget_show ( mainwin->song->ruler );

	gtk_signal_connect ( GTK_OBJECT ( song_get_adjustment ( mainwin->song ) ) , "value_changed" , GTK_SIGNAL_FUNC ( song_update_ruler ) , mainwin->song );


	gtk_box_pack_start ( GTK_BOX ( mainwin->song->rulerwin ) , mainwin->song->ruler , TRUE , TRUE , 0 ) ;

	gtk_box_pack_start ( GTK_BOX ( mainwin->songwin ) , mainwin->song->rulerwin , FALSE , FALSE , 0 );
	gtk_box_pack_start ( GTK_BOX ( mainwin->songwin ) , mainwin->scrollwin , TRUE , TRUE , 0 );
	gtk_box_pack_start ( GTK_BOX ( mainwin->songwin ) , mainwin->song->scrollbar , FALSE , FALSE , 0 );
	gtk_widget_show_all (mainwin->songwin);

	mainwin->paned = gtk_hpaned_new ();
	gtk_paned_add1 (GTK_PANED (mainwin->paned), mainwin->sidebar);
	gtk_paned_add2 (GTK_PANED (mainwin->paned), GTK_WIDGET (mainwin->songwin));

	gtk_paned_set_position (GTK_PANED (mainwin->paned), 
				gnome_config_get_int ("/mixmagic/sidebar/width=150"));


	gnome_app_set_contents (GNOME_APP (mainwin->toplevel), mainwin->paned);

	/* Restore the saved window size */
	default_width = gnome_config_get_int ("/mixmagic/mainwin/width=700");
	default_height = gnome_config_get_int ("/mixmagic/mainwin/height=500");
	gtk_window_set_default_size (GTK_WINDOW (mainwin->toplevel), default_width, default_height);

	gtk_widget_show_all (mainwin->toplevel);

	/* We need to re-size the box to the left of the ruler
	   (which makes the ruler start at the right pixel)
	   after we've shown everything because otherwise the
	   width allocations don't exist yet				

	   Not sure why the -1 in the next line (possibly a line not counted)
	   but it seems to do the trick best					*/

	ruler_pad = ( (TRACK(mainwin->song->track_list->data)->frame->allocation.width - mainwin->song->ruler_label->allocation.width ) / 2 ) + GTK_BOX ( TRACK(mainwin->song->track_list->data))->spacing - 1;

	gtk_box_set_child_packing ( GTK_BOX ( mainwin->song->rulerwin ) , mainwin->song->ruler_label , FALSE , TRUE , ruler_pad , GTK_PACK_START );


	show_sidebar = gnome_config_get_bool ("/mixmagic/sidebar/show=true");

	if (show_sidebar) {
		mainwin_sidebar_show_cb (mainwin->sidebar, mainwin);
	} else {
		mainwin_sidebar_hide_cb (mainwin->sidebar, mainwin);
	}
	gtk_check_menu_item_set_active (GTK_CHECK_MENU_ITEM (view_menu[SIDEBAR_INDEX].widget), show_sidebar);

	return GTK_OBJECT (mainwin);
}

/*
 * This function saves the size of the window
 */
static void
mainwin_size_allocate (GtkWidget *widget, GtkAllocation *allocation, gpointer data) 
{
#if 0
	(* GTK_WIDGET_CLASS (parent_class)->size_allocate) (widget, allocation);
#endif
	gnome_config_set_int ("/mixmagic/mainwin/width", allocation->width);
	gnome_config_set_int ("/mixmagic/mainwin/height", allocation->height);
	gnome_config_sync ();
}


static void 
mainwin_sidebar_show_cb (GtkWidget *widget, Mainwin *mainwin)
{
	gtk_widget_show (widget);

	gtk_paned_set_position (GTK_PANED (mainwin->paned), 
				gnome_config_get_int ("/mixmagic/Sidebar/width=150"));

	gtk_check_menu_item_set_active (GTK_CHECK_MENU_ITEM (view_menu[SIDEBAR_INDEX].widget), TRUE);
}

static void 
mainwin_sidebar_hide_cb (GtkWidget *widget, Mainwin *mainwin)
{
	gtk_widget_hide (widget);

	gtk_paned_set_position (GTK_PANED (mainwin->paned), 0);

	gtk_check_menu_item_set_active (GTK_CHECK_MENU_ITEM (view_menu[SIDEBAR_INDEX].widget), FALSE);
}

static void 
mainwin_sidebar_toggle_cb (GtkWidget *widget, Mainwin *mainwin)
{

	if (GTK_CHECK_MENU_ITEM (widget)->active)
		mainwin_sidebar_show_cb (mainwin->sidebar, mainwin);	
	else
		mainwin_sidebar_hide_cb (mainwin->sidebar, mainwin);	

	gnome_config_set_bool ("/mixmagic/sidebar/show", GTK_CHECK_MENU_ITEM (widget)->active);
	gnome_config_sync ();
}

static gboolean
mainwin_delete_event (GtkWidget *widget, GdkEventAny *event, Mainwin *mainwin)
{
	mainwin_close_clicked (NULL, mainwin);

	return TRUE;
}

/*
 * Close the window.
 */
static void
mainwin_close_clicked (GtkWidget *widget, Mainwin *mainwin)
{
	g_assert (mainwin);
	g_assert (IS_MAINWIN (mainwin));
	
	if (mainwin_close_if_user_permits (mainwin) == FALSE)
		return;

	window_list = g_list_remove (window_list, mainwin);

	mainwin_stop_clicked (NULL, mainwin);

	gtk_object_destroy (GTK_OBJECT (mainwin));

	/* 
	 *  Terminate the app if this is the last instance 
	 */
	if (g_list_length (window_list) == 0)
		gtk_main_quit ();
}

/*
 * Zoom In.
 */
static void
mainwin_zoom_in_clicked (GtkWidget *widget, Mainwin *mainwin)
{
	song_zoom_in (mainwin->song);
}

/*
 * Zoom out.
 */
static void
mainwin_zoom_out_clicked (GtkWidget *widget, Mainwin *mainwin)
{
	song_zoom_out (mainwin->song);
}

/*
 * Zoom fit
 */
static void
mainwin_zoom_fit_clicked (GtkWidget *widget, Mainwin *mainwin)
{
	song_zoom_fit (mainwin->song);
}

/*
 * Song information
 */
static void
mainwin_song_info_clicked (GtkWidget *widget, Mainwin *mainwin)
{
	g_warning("Function not completed yet!");
}

/*
 * Exit the app.
 */
static void
mainwin_exit_clicked (GtkWidget *widget, gpointer data)
{
	Mainwin *mainwin = MAINWIN (data);

	song_stop_playback (mainwin->song);
	gtk_main_quit ();
}

/*
 * User clicked 'New'.
 */
static void
mainwin_new_clicked (GtkWidget *widget, gpointer data)
{
	mainwin_new (NULL);
}

/*
 * User clicked 'Preferences'.
 */
static void
mainwin_preferences_clicked (GtkWidget *widget, gpointer data)
{
	Mainwin *mainwin = MAINWIN (data);

	/* Stop the playback to avoid segfaults when changing output plugins */
	song_stop_playback (mainwin->song);

	preferences_show (mainwin->toplevel);
}

/**
 * mainwin_close_if_user_permits : If the workbook is dirty the user is
 *              prompted to see if they should exit.
 *
 * Returns : TRUE is the book remains open.
 *           FALSE if it is closed.
 * (Stolen from gnumeric)
 */
static gboolean
mainwin_close_if_user_permits (Mainwin *mw)
{
	gboolean   can_close = TRUE;
	gboolean   done      = FALSE;
	int        iteration = 0;
	static int in_can_close = FALSE;
	
	if (in_can_close)
		return FALSE;
	in_can_close = TRUE;
	
	while (song_is_dirty (mw->song) && !done) {
		
		GtkWidget *d, *l, *cancel_button;
		int button;
		char *s;
		
		iteration++;
		
		d = gnome_dialog_new ("Warning",
				      GNOME_STOCK_BUTTON_YES,
				      GNOME_STOCK_BUTTON_NO,
				      GNOME_STOCK_BUTTON_CANCEL,
				      NULL);
		cancel_button = g_list_last (GNOME_DIALOG (d)->buttons)->data;
		gtk_widget_grab_focus (cancel_button);
		gnome_dialog_set_parent (GNOME_DIALOG (d), GTK_WINDOW (mw->toplevel));
		
		if (mw->filename)
			s = g_strdup_printf (_("%s has unsaved changes, save them?"),
					     g_basename (mw->filename));
		else
			s = g_strdup (_("Song has unsaved changes, save them?"));
		
		l = gtk_label_new (s);
		gtk_widget_show (l);
		g_free (s);
		
		gtk_box_pack_start (GTK_BOX (GNOME_DIALOG (d)->vbox), l, TRUE, TRUE, 0);
		
		gtk_window_set_position (GTK_WINDOW (d), GTK_WIN_POS_MOUSE);
		button = gnome_dialog_run_and_close (GNOME_DIALOG (d));
		
		switch (button) {
			/* YES */
		case 0:
			done = mainwin_save (NULL, mw);
			break;
			
			/* NO */
		case 1:
			can_close = TRUE;
			done      = TRUE;
			break;
			
			/* CANCEL */
		case -1:
		case 2:
			can_close = FALSE;
			done      = TRUE;
			break;
		}
	}
	
	in_can_close = FALSE;
	
	return can_close;
}

/*
 * sets the *flag to true and exits the gtk mainloop.
 */
static void
set_true_and_quit (GtkWidget *widget, gboolean *flag)
{
	*flag = TRUE;

	gtk_main_quit ();
}


/*
 * User clicked 'About'.
 */
static void
mainwin_about_clicked (GtkWidget *widget, gpointer data)
{
	g_assert(data != NULL);
	g_assert(IS_MAINWIN(data));

	about_show(MAINWIN(data)->toplevel);
}

/*
 * Code to create and run the open-dialog, takes a pointer to a string and returns TRUE if
 * user clicked ok
 */
static gboolean
mainwin_show_open_dlg (Mainwin *mainwin, gchar **filename)
{
	gboolean ok_clicked = FALSE;

	g_assert (mainwin != NULL);
	g_assert (IS_MAINWIN(mainwin));
	g_assert (filename != NULL);

	if (mainwin->open_file_sel == NULL) {

		mainwin->open_file_sel = gtk_file_selection_new (_("Open song (*.mmsong)"));
		
		gtk_signal_connect (GTK_OBJECT (GTK_FILE_SELECTION (mainwin->open_file_sel)->ok_button), "clicked",
				    GTK_SIGNAL_FUNC (set_true_and_quit), &ok_clicked);
		
		gtk_signal_connect (GTK_OBJECT (GTK_FILE_SELECTION (mainwin->open_file_sel)->cancel_button), "clicked",
				    GTK_SIGNAL_FUNC (gtk_main_quit), NULL);
		gtk_signal_connect (GTK_OBJECT (mainwin->open_file_sel), "delete_event",
						GTK_SIGNAL_FUNC (gtk_main_quit), NULL);

	}
	gtk_widget_show (mainwin->open_file_sel);
	/* Run the dialog */
	gtk_main ();
	gtk_widget_hide (mainwin->open_file_sel);
	
	if (ok_clicked) {
		*filename = g_strdup (gtk_file_selection_get_filename (GTK_FILE_SELECTION (mainwin->open_file_sel)));
	} else
		*filename = NULL;

	return ok_clicked;
}

/*
 * User clicked 'Import song'
 */
static void
mainwin_import_clicked (GtkWidget *widget, gpointer data)
{
	Mainwin *mainwin = (Mainwin *)data;
	gchar *filename = NULL;

	g_assert (mainwin != NULL);
	g_assert (IS_MAINWIN(mainwin));

	if (mainwin_show_open_dlg(mainwin, &filename) && filename != NULL) {
		if (!song_load (mainwin->song, filename)) {
			gchar *error;

			error = g_strdup_printf (_("Failed to open '%s'.\nIs it a .mmsong file?"),
						 filename);
			
			gnome_error_dialog_parented (error, GTK_WINDOW(mainwin->toplevel));
			g_free (error);
		}
		if (filename)
			g_free (filename);
	}
}


/*
 * User clicked 'Open'.
 */
static void
mainwin_open_clicked (GtkWidget *widget, gpointer data)
{
	Mainwin *mainwin = (Mainwin *)data;
	gchar *filename = NULL;

	g_assert (mainwin != NULL);

	if (mainwin_show_open_dlg(mainwin, &filename)) {
		Mainwin *new_win;
		gchar *title;
		
		if (song_is_empty (mainwin->song)) {

			if (mainwin->filename)
				g_free (mainwin->filename);
			
			mainwin->filename = filename;
			
			if (song_load (mainwin->song, filename)) {
				
				title = g_strdup_printf (_("MixMagic: %s"), g_basename (filename));
				gtk_window_set_title (GTK_WINDOW (mainwin->toplevel), title);
				g_free (title);
				
				song_zoom_fit (mainwin->song);
			} else {
				gchar *error;

				error = g_strdup_printf (_("Failed to open '%s'.\nIs it a .mmsong file?"),
							 filename);

				gnome_error_dialog_parented (error, GTK_WINDOW(mainwin->toplevel));
				g_free (error);
				
			}
		} else {
			
			new_win = MAINWIN (mainwin_new (filename));
			
			g_free (filename);
		}
		
	}
}

/*
 * User clicked 'save'.
 */
gboolean
mainwin_save (GtkWidget *widget, gpointer data)
{
	Mainwin *mainwin = (Mainwin *)data;
	
	if (mainwin->filename)
		return song_save (mainwin->song, mainwin->filename);
	else 
		return mainwin_save_as (NULL, mainwin);
}

/*
 * User clicked 'save'.
 */
gboolean
mainwin_save_as (GtkWidget *widget, gpointer data)
{
	Mainwin *mainwin = (Mainwin *)data;
	gboolean ok_clicked = FALSE;
	gchar *title;

	if (mainwin->save_file_sel == NULL) {
		
		mainwin->save_file_sel = gtk_file_selection_new (_("Save song (*.mmsong)"));
		
		gtk_signal_connect (GTK_OBJECT (GTK_FILE_SELECTION (mainwin->save_file_sel)->ok_button), "clicked",
				    GTK_SIGNAL_FUNC (set_true_and_quit), &ok_clicked);
		
		gtk_signal_connect (GTK_OBJECT (GTK_FILE_SELECTION (mainwin->save_file_sel)->cancel_button), "clicked",
				    GTK_SIGNAL_FUNC (gtk_main_quit), NULL);
		gtk_signal_connect (GTK_OBJECT (mainwin->save_file_sel), "delete_event",
						GTK_SIGNAL_FUNC (gtk_main_quit), NULL);
		
	}
	gtk_widget_show (mainwin->save_file_sel);
	/* Run the dialog */
	gtk_main ();
	gtk_widget_hide (mainwin->save_file_sel);

	if (ok_clicked == FALSE)
		return FALSE;

	/* Free the old filename */
	if (mainwin->filename)
		g_free (mainwin->filename);

	mainwin->filename = g_strdup (gtk_file_selection_get_filename (GTK_FILE_SELECTION (mainwin->save_file_sel)));
	
	if (song_save (mainwin->song, mainwin->filename)) {

		title = g_strdup_printf (_("MixMagic: %s"), g_basename (mainwin->filename));
		gtk_window_set_title (GTK_WINDOW (mainwin->toplevel), title);
		g_free (title);
	} else {
		gchar *error;
			
		error = g_strdup_printf (_("Failed to save '%s'."),
					 mainwin->filename);
		gnome_error_dialog_parented (error, GTK_WINDOW(mainwin->toplevel));
		g_free (error);			
		
		return FALSE;
	}
	return TRUE;
}

/*
 * Signal handler called when the selections owner returns the data 
 */
static void 
mainwin_selection_received (GtkWidget *widget, GtkSelectionData *selection_data, guint time, gpointer data)
{
	Block *block;
	Mainwin *mainwin = MAINWIN (data);

	/* Is it a MMBlock selection */
	if (selection_data->type != block_atom) {

		/* Start playback from the beginning */
		song_play (mainwin->song);
		return;
	}

	block = *(Block **)selection_data->data;

	/*
	 * Block selected, just play the selection.
	 * The block gets deleted by the mixer.
	 */
	
	song_play_block_and_delete (mainwin->song, block);	
}

/*
 * User clicked 'Rewind'.
 */
static void
mainwin_rewind_clicked (GtkWidget *widget, Mainwin *mainwin)
{
	song_stop_playback (mainwin->song);
	song_move_marker (mainwin->song, 0);

	gtk_adjustment_set_value (song_get_adjustment (mainwin->song), 0);
}

/*
 * User clicked 'play'.
 */
void
mainwin_play_clicked (GtkWidget *widget, Mainwin *mainwin)
{
        /* Get the atom corresponding to the string "Block" */
        if (block_atom == GDK_NONE)
		block_atom = gdk_atom_intern ("Block", FALSE);

	/* If we are allready playing, then stop */
	if (mixer_is_playing ())
		song_stop_playback (mainwin->song);
	else {
		/* Else Check if a block is selected */

		/* Tell mainwin_selection_received to play the selection */
		gtk_selection_convert (mainwin->toplevel, GDK_SELECTION_PRIMARY, 
				       block_atom,
				       GDK_CURRENT_TIME);
	}
}

/*
 * User clicked 'stop'.
 */
void
mainwin_stop_clicked (GtkWidget *widget, Mainwin *mainwin)
{
	song_stop_playback (mainwin->song);
}


/*
 * User toggled 'Repeat'.
 */
static void
mainwin_repeat_toggled (GtkWidget *widget, Mainwin *mainwin)
{
	mainwin->song->repeat_flag = gtk_toggle_button_get_active(GTK_TOGGLE_BUTTON(widget));
}
