/* -*- Mode: C; tab-width: 8; indent-tabs-mode: t; c-basic-offset: 8 -*- */
/*  This file is part of MixMagic.

    Copyright (C) 2000-2001 Jonas Borgstrm <jonas@bengburken.net>

    This program is free software; you can redistribute it and/or
    modify it under the terms of the GNU General Public
    License as published by the Free Software Foundation; either
    version 2 of the License, or (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the 
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software 
    Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
*/

#include <math.h>
#include "mixmagic.h"
#include "track.h"
#include "block.h"
#include "loopblock.h"
#include "mixer.h"

static GtkHBoxClass *parent_class = NULL;
static GdkAtom block_atom = GDK_NONE;
static GtkWidget *insert_file_sel = NULL;

static void track_finalize (GtkObject *object);
static void track_class_init (TrackClass *class);
static void track_init (Track *track);
static gint track_button_press_event (GtkWidget *widget, GdkEventButton *event, Track *track);
static gint track_button_release_event (GtkWidget *widget, GdkEventButton *event, Track *track);
static void track_selection_received (GtkWidget *widget, GtkSelectionData *selection_data, guint time, Track *track);
static void track_paste_clicked (GtkWidget *widget, Track *track);
static void track_paste_loopblock_clicked (GtkWidget *widget, Track *track);
static void track_insert_sample_clicked (GtkWidget *widget, Track *track);
static void track_solo_toggled (GtkWidget *widget, Track *track);
static void track_mute_toggled (GtkWidget *widget, Track *track);
static void track_amp_changed (GtkAdjustment *adj, Track *track);
static void track_pan_changed (GtkAdjustment *adj, Track *track);
static gboolean track_drag_data_received (GtkWidget *widget, GdkDragContext *drag_context, gint x, gint y, GtkSelectionData *selection_data, guint info, guint time, Track *track);

guint
track_get_type ()
{
        static guint track_type = 0;
	
        if (!track_type) {

		GtkTypeInfo track_info = {
			"Track",
			sizeof (Track),
			sizeof (TrackClass),
			(GtkClassInitFunc) track_class_init,
			(GtkObjectInitFunc) track_init,
			(GtkArgSetFunc) NULL,
			(GtkArgGetFunc) NULL
		};
	    
		track_type = gtk_type_unique (gtk_hbox_get_type (), &track_info);
          }
	
        return track_type;
}

static void
track_class_init (TrackClass *class)
{
	GtkWidgetClass *widget_class;
	GtkObjectClass *object_class;

	parent_class = gtk_type_class (GTK_TYPE_HBOX);

	object_class = (GtkObjectClass*) class;
	object_class->finalize = track_finalize;

	widget_class = (GtkWidgetClass*) class;
}

static void
track_init (Track *track)
{
	GtkWidget *item, *separator, *table, *label;
	GtkWidget *hbox;
	gint index;
	GtkStyle *layout_style;

	static GtkTargetEntry target_table[] = {
		{ "STRING",     0, TARGET_STRING },
		{ "text/plain", 0, TARGET_STRING },
	};

	gtk_box_set_spacing (GTK_BOX(track), 5);
	track->frame = gtk_frame_new (_("Track X"));

	table = gtk_table_new (4, 2, FALSE);

	gtk_container_add (GTK_CONTAINER (track->frame), table);

	hbox = gtk_hbox_new(TRUE, 0);
	gtk_table_attach_defaults (GTK_TABLE (table), hbox, 0, 2, 0, 1);

	track->solo_toggle_button = gtk_toggle_button_new_with_label (_("Solo"));
	track->mute_toggle_button = gtk_toggle_button_new_with_label (_("Mute"));
	gtk_signal_connect (GTK_OBJECT (track->solo_toggle_button), "toggled",
			    GTK_SIGNAL_FUNC (track_solo_toggled), track);
	gtk_signal_connect (GTK_OBJECT (track->mute_toggle_button), "toggled",
			    GTK_SIGNAL_FUNC (track_mute_toggled), track);

	gtk_box_pack_start(GTK_BOX(hbox), track->solo_toggle_button, FALSE, TRUE, GNOME_PAD_SMALL);
	gtk_box_pack_start(GTK_BOX(hbox), track->mute_toggle_button, FALSE, TRUE, GNOME_PAD_SMALL);

	track->amp = 1;
	track->amp_adj = GTK_ADJUSTMENT (gtk_adjustment_new (0, -200, 10, 0.01, 0.01, 0.0));

	gtk_signal_connect (GTK_OBJECT (track->amp_adj), "value_changed",
			    GTK_SIGNAL_FUNC (track_amp_changed), track);

	track->pan = 0;
	track->pan_adj = GTK_ADJUSTMENT (gtk_adjustment_new (track->pan, -1, 1, 0.01, 0.01, 0.0));

	gtk_signal_connect (GTK_OBJECT (track->pan_adj), "value_changed",
			    GTK_SIGNAL_FUNC (track_pan_changed), track);

	label = gtk_label_new (_("dB:"));
	gtk_table_attach_defaults (GTK_TABLE (table), label, 0, 1, 2, 3);
	track->amp_spin_button = gtk_spin_button_new (track->amp_adj, 1, 2);
	gtk_widget_set_usize (track->amp_spin_button, 
			      gdk_string_width (track->amp_spin_button->style->font, "-10,000000"),
			      -1);

	gtk_table_attach_defaults (GTK_TABLE (table), track->amp_spin_button, 1, 2, 2, 3);

	label = gtk_label_new (_("Pan:"));
	gtk_table_attach_defaults (GTK_TABLE (table), label, 0, 1, 3, 4);

	track->pan_hscale = gtk_hscale_new(track->pan_adj);
	gtk_scale_set_value_pos(GTK_SCALE(track->pan_hscale), GTK_POS_RIGHT);
	gtk_table_attach_defaults (GTK_TABLE (table), track->pan_hscale, 1, 2, 3, 4);

	track->layout = gtk_layout_new (NULL, NULL);

	/* Make layout background dark */

	layout_style = gtk_style_copy ( gtk_widget_get_style (GTK_WIDGET(track->layout)));

	for ( index=0 ; index < 5 ; index++ )
		layout_style->base[index] = layout_style->dark[index];

	gtk_widget_set_style (GTK_WIDGET(track->layout) , layout_style);


	gtk_widget_set_usize (GTK_WIDGET(track->layout), -1, 100); /* Layout-height=100 */

#if 1 /* Fix me!!!!!! Why???? */
	gtk_object_ref (GTK_OBJECT (track->layout));
#endif
	/* Enable dropping */
	gtk_drag_dest_set (track->layout, GTK_DEST_DEFAULT_MOTION | GTK_DEST_DEFAULT_DROP, 
			   target_table, 2, GDK_ACTION_COPY);
	gtk_signal_connect (GTK_OBJECT (track->layout), "drag_data_received",
			    GTK_SIGNAL_FUNC (track_drag_data_received), track);	

	track->marker = gtk_event_box_new ();

	separator = gtk_vseparator_new ();

	gtk_container_add (GTK_CONTAINER (track->marker), separator);
	gtk_widget_set_usize (track->marker, 1, 200);
	gtk_layout_put (GTK_LAYOUT (track->layout), track->marker, 0, 0);

	gtk_widget_set_events (track->layout, GDK_BUTTON_PRESS_MASK | GDK_BUTTON_RELEASE_MASK);

	gtk_signal_connect (GTK_OBJECT (track->layout), "button_press_event",
			    GTK_SIGNAL_FUNC (track_button_press_event), track);

	gtk_signal_connect (GTK_OBJECT (track->layout), "button_release_event",
			    GTK_SIGNAL_FUNC (track_button_release_event), track);

	gtk_signal_connect (GTK_OBJECT (track->layout), "selection_received",
			    GTK_SIGNAL_FUNC (track_selection_received), track);

	gtk_box_pack_start (GTK_BOX (track), 
			    track->frame,
			    FALSE, FALSE, 0);

	gtk_box_pack_start (GTK_BOX (track), 
			    track->layout,
			    TRUE, TRUE, 0);

	gtk_widget_show_all (GTK_WIDGET (track));

	/* Create the popup menu */
        track->popup_menu = gtk_menu_new();

        item = gtk_menu_item_new_with_label(_("Paste"));
        gtk_menu_append(GTK_MENU(track->popup_menu), item);
        gtk_widget_show(item);
	gtk_signal_connect (GTK_OBJECT (item), "activate",
			    GTK_SIGNAL_FUNC (track_paste_clicked), track);

        item = gtk_menu_item_new_with_label(_("Paste loopblock"));
        gtk_menu_append(GTK_MENU(track->popup_menu), item);
        gtk_widget_show(item);
	gtk_signal_connect (GTK_OBJECT (item), "activate",
			    GTK_SIGNAL_FUNC (track_paste_loopblock_clicked), track);
	/* Separator */
        item = gtk_menu_item_new ();
        gtk_menu_append(GTK_MENU(track->popup_menu), item);
        gtk_widget_show(item);

        item = gtk_menu_item_new_with_label(_("Insert sample"));
        gtk_menu_append(GTK_MENU(track->popup_menu), item);
        gtk_widget_show(item);
	gtk_signal_connect (GTK_OBJECT (item), "activate",
			    GTK_SIGNAL_FUNC (track_insert_sample_clicked), track);


}

void
track_set_mute (Track *track, gboolean mute)
{
	gtk_toggle_button_set_active (GTK_TOGGLE_BUTTON (track->mute_toggle_button), mute);
}

void
track_set_amp (Track *track, gfloat amp)
{
	track->amp = amp;
	/* Convert it from dB */
	track->amp_adj->value = 10 * log10 (amp);

	gtk_adjustment_value_changed (track->amp_adj);
}

static void
track_amp_changed (GtkAdjustment *adj, Track *track)
{
	track->amp = pow (10, adj->value / 10.0);

	gtk_widget_grab_focus (track->solo_toggle_button);
}

void
track_set_pan (Track *track, gfloat pan)
{
	track->pan = pan;
	track->pan_adj->value = pan;

	gtk_adjustment_value_changed (track->pan_adj);
}

static void
track_pan_changed (GtkAdjustment *adj, Track *track)
{
	track->pan = adj->value;

	gtk_widget_grab_focus (track->solo_toggle_button);
}

static void
mute_track (gpointer data, gpointer user_data)
{
	Track *track = TRACK (data);
	gtk_toggle_button_set_active (GTK_TOGGLE_BUTTON (track->mute_toggle_button), TRUE);
}

static void
unmute_track (gpointer data, gpointer user_data)
{
	Track *track = TRACK (data);

	gtk_toggle_button_set_active (GTK_TOGGLE_BUTTON (track->mute_toggle_button), FALSE);
}

static void
unsolo_track (gpointer data, gpointer user_data)
{
	Track *track = TRACK (data);

	if (track->going_solo_flag == FALSE)
		gtk_toggle_button_set_active (GTK_TOGGLE_BUTTON (track->solo_toggle_button), FALSE);
}

static void
track_solo_toggled (GtkWidget *widget, Track *track)
{
	if (GTK_TOGGLE_BUTTON (widget)->active) {
		/* Set this flag to avoid infinite loops */
		track->going_solo_flag = TRUE;
		g_list_foreach (track->song->track_list, unsolo_track, NULL);
		g_list_foreach (track->song->track_list, mute_track, NULL);
		gtk_toggle_button_set_active (GTK_TOGGLE_BUTTON (track->mute_toggle_button), FALSE);
		track->going_solo_flag = FALSE;

	} else if (track->song->going_unsolo_flag == FALSE){
		g_list_foreach (track->song->track_list, unmute_track, NULL);
	}
}

static gboolean 
track_drag_data_received (GtkWidget *widget, GdkDragContext *drag_context, gint x, gint y, 
			  GtkSelectionData *selection_data, guint info, guint time, Track *track)
{
	Block *block;
	Sample *sample;
	GList *node;

	for (node = gnome_uri_list_extract_filenames(selection_data->data);node;node = node->next) {

		g_assert(node->data != NULL);
		g_assert (track != NULL);

		g_message("%s", (gchar *)node->data);

		sample = song_get_sample (track->song, (gchar *)node->data) ;

		if (sample == NULL) /* failed to load, I'd better bail out now... */
			return FALSE;

		block = BLOCK (block_new (track));
		block->sample = sample_ref (sample);

		block->start = 0;
		block->length = sample_get_length(block->sample);

		/* Position it on the same place as the popup menu */
		block->position = (x + (gint)(GTK_LAYOUT (track->layout)->hadjustment->value)) * track->song->x_scale;

		/* Insert the block into the song */
		song_insert_block (track->song, block, track->track_num);

		if (node->next) /* Better change tracks */
			track = TRACK(g_list_last(track->song->track_list)->data);
	}

	return FALSE; 
}

static void
track_mute_toggled (GtkWidget *widget, Track *track)
{
	if (GTK_TOGGLE_BUTTON (track->solo_toggle_button)->active &&
	    GTK_TOGGLE_BUTTON (widget)->active == TRUE &&
	    track->going_solo_flag == FALSE ) {

		track->song->going_unsolo_flag = TRUE;
		gtk_toggle_button_set_active (GTK_TOGGLE_BUTTON (track->solo_toggle_button), FALSE);
		track->song->going_unsolo_flag = FALSE;
	}

	if ((track->mute_flag = GTK_TOGGLE_BUTTON (widget)->active) == FALSE &&
	    track->going_solo_flag == FALSE ) {

		track->song->going_unsolo_flag = TRUE;
		g_list_foreach (track->song->track_list, unsolo_track, NULL);
		track->song->going_unsolo_flag = FALSE;
	}
}

/*
 * Handle button press events 
 */
static gint
track_button_press_event (GtkWidget *widget, GdkEventButton *event, Track *track)
{
	/* Display the popup menu, when the user
	   clicks on the right mouse button */

	if (event->button == 3) {

		track->menu_point.x = event->x;
		track->menu_point.y = event->y;

		gtk_menu_popup (GTK_MENU (track->popup_menu), NULL, NULL, NULL, NULL,
				event->button, event->time);
		return TRUE;
	}
	return FALSE;
}

/*
 * Handle button release events 
 */
static gint
track_button_release_event (GtkWidget *widget, GdkEventButton *event, Track *track)
{
	gint x, y;
        GdkModifierType mods;

	if (event->button == 1) {

		if (event->window != widget->window)
			gdk_window_get_pointer (widget->window, &x, &y, &mods);

		song_move_marker (track->song, (x + 
					 (gint)(GTK_LAYOUT (track->layout)->hadjustment->value)) * track->song->x_scale);

		return TRUE;
	}
	return FALSE;
}

/*
 * Called when the use clicks on the "play" button in the fileselctor.
 */
static void
track_insert_sample_play_clicked (GtkWidget *widget, Track *track)
{
	/*
	 * Play the sample
	 */
	Sample *sample;
	
	sample = song_get_sample (NULL, 
				  gtk_file_selection_get_filename (GTK_FILE_SELECTION (insert_file_sel)));
	
	if (sample)
		mixer_play_sample_and_delete (sample);
}

/*
 * Called when the use clicks on the "ok" button in the fileselector.
 */
static void
track_insert_sample_ok_clicked (GtkWidget *widget, Track *track)
{
	Block *block;
	Sample *sample;

	g_assert (track != NULL);

	sample = song_get_sample (track->song, 
				  gtk_file_selection_get_filename (GTK_FILE_SELECTION (insert_file_sel)));
	
	if (sample == NULL)
		return;

	block = BLOCK (block_new (track));

	block->sample = sample_ref (sample);
	block->start = 0;
	block->length = block->sample->length / block->sample->byte_p_spl;

	/* Position it on the same place as the popup menu */
	block->position = (track->menu_point.x + 
			   (gint)(GTK_LAYOUT (track->layout)->hadjustment->value)) * track->song->x_scale;

	/* Insert the block into the song */
	song_insert_block (track->song, block, track->track_num);

	/* Hide the file selection dialog */
	gtk_widget_hide (insert_file_sel);
}

/*
 * Called when the use clicks on the "cancel" button in a GtkFileselector.
 */
static void
track_file_sel_cancel_clicked (GtkWidget *widget, gpointer data)
{
	GtkWidget *file_sel = (GtkWidget *)data;

	g_assert (file_sel != NULL);
	
	gtk_widget_hide (file_sel);
}

/*
 * Displays the "Insert sample" file_selection dialog.
 */
static void
track_insert_sample_clicked (GtkWidget *widget, Track *track)
{
	static gint sig_id = -1;
	g_assert (track != NULL);
	
	if (insert_file_sel == NULL) {
		GtkWidget *play;

		insert_file_sel = gtk_file_selection_new (_("Insert sample (*.wav)"));

		/* Create a play button */
		play = gtk_button_new_with_label (_("Play"));
		gtk_widget_show (play);
		gtk_signal_connect (GTK_OBJECT (play), "clicked",
			    GTK_SIGNAL_FUNC (track_insert_sample_play_clicked), track);

		/* Add it to the fileselector */
		gtk_box_pack_start (GTK_BOX (GTK_FILE_SELECTION (insert_file_sel)->button_area), play, FALSE, FALSE, 0);
		
		sig_id = gtk_signal_connect (GTK_OBJECT (GTK_FILE_SELECTION (insert_file_sel)->ok_button), "clicked",
			    GTK_SIGNAL_FUNC (track_insert_sample_ok_clicked), track);

		gtk_signal_connect (GTK_OBJECT (GTK_FILE_SELECTION (insert_file_sel)->cancel_button), "clicked",
			    GTK_SIGNAL_FUNC (track_file_sel_cancel_clicked), insert_file_sel);

	} else {
		/* Update the signal handler to point to the correct object */
		gtk_signal_disconnect (GTK_OBJECT (GTK_FILE_SELECTION (insert_file_sel)->ok_button), sig_id);
		sig_id = gtk_signal_connect (GTK_OBJECT (GTK_FILE_SELECTION (insert_file_sel)->ok_button), "clicked",
					     GTK_SIGNAL_FUNC (track_insert_sample_ok_clicked), track);
		
	}
	gtk_widget_show (insert_file_sel);
}

/*
 * User clicked 'Paste'.
 */
static void
track_paste_clicked (GtkWidget *widget, Track *track)
{
	
        /* Get the atom corresponding to the string "MMBlock" */
        if (block_atom == GDK_NONE)
		block_atom = gdk_atom_intern ("Block", FALSE);

	track->paste_mode = PASTE_BLOCK;

        /* And request the "MMblock" target for the primary selection */
        gtk_selection_convert (track->layout, GDK_SELECTION_PRIMARY, block_atom,
				GDK_CURRENT_TIME);
}

/*
 * User clicked 'Paste loopblock'.
 */
static void
track_paste_loopblock_clicked (GtkWidget *widget, Track *track)
{
	
        /* Get the atom corresponding to the string "MMBlock" */
        if (block_atom == GDK_NONE)
		block_atom = gdk_atom_intern ("Block", FALSE);

	track->paste_mode = PASTE_LOOPBLOCK;

        /* And request the "MMblock" target for the primary selection */
        gtk_selection_convert (track->layout, GDK_SELECTION_PRIMARY, block_atom,
				GDK_CURRENT_TIME);
}

/*
 * Signal handler called when the selections owner returns the data 
 */
static void 
track_selection_received (GtkWidget *widget, GtkSelectionData *selection_data, guint time, Track *track)
{
	Block *block;
	LoopBlock *lblock;

	/* Is it a MMBlock selection */
	if (selection_data->type != block_atom)
		return;

	block = *(Block **)selection_data->data;

	switch (track->paste_mode) {

	case PASTE_BLOCK:
		/* Position it on the same place as the popup menu */
		block->position = (track->menu_point.x + 
				   (gint)(GTK_LAYOUT (track->layout)->hadjustment->value)) * track->song->x_scale;
		
		/* Insert the block into the song */
		song_insert_block (track->song, block, track->track_num);
		break;

	case PASTE_LOOPBLOCK:

		lblock = LOOPBLOCK (loopblock_new_from_block (block, track));
		/* We don't need this block anymore */
		gtk_object_unref (GTK_OBJECT (block));

		/* Position it on the same place as the popup menu */
		lblock->block.position = (track->menu_point.x + 
				   (gint)(GTK_LAYOUT (track->layout)->hadjustment->value)) * track->song->x_scale;

		/* Insert the loopblock into the song */
		song_insert_block (track->song, BLOCK (lblock), track->track_num);
		break;
	};

}

/*
 * Creates a new track.
 */
GtkWidget *
track_new (GtkAdjustment *adjustment, Song *song, gint track_num)
{
	gchar buffer[20];
        Track *track =  TRACK (gtk_type_new (track_get_type ()));

	track_set_adjustment (track, adjustment);
	track->song = song;
	track->track_num = track_num;

	sprintf (buffer, _("Track %d"), track_num);
	gtk_frame_set_label (GTK_FRAME (track->frame), buffer);

	return GTK_WIDGET (track);
}

GtkAdjustment*
track_get_adjustment (const Track *track)
{
        g_return_val_if_fail (track != NULL, NULL);
        g_return_val_if_fail (IS_TRACK (track), NULL);
	
        return gtk_layout_get_hadjustment (GTK_LAYOUT (track->layout));
}

void
track_set_adjustment (Track *track,
		      GtkAdjustment *adjustment)
{
        g_return_if_fail (track != NULL);
        g_return_if_fail (IS_TRACK (track));

        gtk_layout_set_hadjustment (GTK_LAYOUT (track->layout), adjustment);
}

static void 
track_finalize (GtkObject *object)
{
	Track *track = TRACK (object);

	g_print ("Track finalize\n");

	g_list_free (track->block_list);

	if (track->popup_menu)
		gtk_widget_destroy (track->popup_menu);

	(*GTK_OBJECT_CLASS (parent_class)->finalize) (object);
}

void
track_remove_block (Track *track, Block *block)
{
	gtk_container_remove (GTK_CONTAINER (track->layout), GTK_WIDGET (block));
	track->block_list = g_list_remove (track->block_list, block);
}

void
track_add_block (Track *track, Block *block)
{
	GList *list;
	gint position = 0;

	/* Update the blocks track reference */
	block->track = track;

	/* Go to the correct position */
	list = track->block_list;
	while (list && BLOCK(list->data)->position < block->position) {
		position++;
		list = g_list_next (list);
	}

	if (list == NULL)
		track->block_list = g_list_append (track->block_list, block);
	else
		track->block_list = g_list_insert (track->block_list, block, position);

	/* Add it to the screen */
	gtk_layout_put (GTK_LAYOUT (track->layout), GTK_WIDGET (block), 0, 0);
	gtk_widget_show (GTK_WIDGET (block));
	/* Let the block move to the correct position */

	block_update (block);
}

static void
update_block (gpointer data, gpointer user_data)
{
	block_update (BLOCK (data));
}

void
track_update (Track *track)
{
	g_list_foreach (track->block_list, update_block, NULL);
}

typedef struct _LayoutChild LayoutChild;
struct _LayoutChild {
  GtkWidget *widget;
  gint x;
  gint y;
};

/*
 * This function will sort the gtklayout children.
 */
static gint
child_x_sort (gconstpointer a, gconstpointer b)
{
	LayoutChild *childa = (LayoutChild *)a;
	LayoutChild *childb = (LayoutChild *)b;

	if (childa->x < childb->x)
		return -1;
	else if (childa->x > childb->x)
		return 1;
	else 
		return 0;
}

/*
 * This function raises the widgets gdkwindow.
 */
static void
raise_widget (GtkWidget *widget, gpointer data)
{
	if (widget->window)
		gdk_window_raise (widget->window);
}

static gint
track_block_pos_cmp (gconstpointer b1, gconstpointer b2)
{
	g_assert (IS_BLOCK(b1));
	g_assert (IS_BLOCK(b2));

	if (BLOCK(b1)->position > BLOCK(b2)->position)
		return 1;

	if (BLOCK(b1)->position < BLOCK(b2)->position)
		return -1;

	return 0;
}


/*
 * This vill raise all blocks from the left to the right.
 */
void
track_sort_blocks (Track *track)
{
	GTK_LAYOUT (track->layout)->children = g_list_sort (GTK_LAYOUT (track->layout)->children, 
							    child_x_sort);

	gtk_container_foreach (GTK_CONTAINER (track->layout), raise_widget, NULL);

	track->block_list = g_list_sort(track->block_list, track_block_pos_cmp);

	g_list_foreach (track->block_list, update_block, NULL);
}

void 
track_move_marker (Track *track, gint position)
{
	gtk_layout_move (GTK_LAYOUT (track->layout),
			 track->marker,
			 position, 0);

	gdk_window_raise (track->marker->window);
}

void 
track_show_marker (Track *track)
{
	gtk_widget_show (track->marker);
}

void 
track_hide_marker (Track *track)
{
	gtk_widget_hide (track->marker);

	gtk_layout_move (GTK_LAYOUT (track->layout),
			 track->marker,
			 0, 0);
}

void 
track_set_length (Track *track, gint length)
{
	gtk_layout_set_size (GTK_LAYOUT (track->layout), length, BLOCK_HEIGHT);
}
